""" Genes """
import threading
import sys
import numpy as np

from AnyQt.QtWidgets import (
    QSplitter, QTableView,  QHeaderView, QAbstractItemView, QStyle
)
from AnyQt.QtCore import (
    Qt, QSize, QThreadPool, QAbstractTableModel, QVariant, QModelIndex, QTimer,
)
from AnyQt.QtGui import (
    QFont, QColor
)

from Orange.widgets.gui import (
    vBox, comboBox, ProgressBar, widgetBox, auto_commit, widgetLabel, checkBox,
    rubber, lineEdit, LinkRole, LinkStyledItemDelegate
)
from Orange.widgets.widget import OWWidget
from Orange.widgets.utils import itemmodels
from Orange.widgets.settings import Setting, DomainContextHandler, ContextSetting
from Orange.widgets.utils.signals import Output, Input
from Orange.data import StringVariable, DiscreteVariable, Domain, Table, filter as table_filter


from orangecontrib.bioinformatics.widgets.utils.data import (
    TAX_ID, GENE_AS_ATTRIBUTE_NAME, GENE_ID_COLUMN, GENE_ID_ATTRIBUTE
)
from orangecontrib.bioinformatics.widgets.utils.concurrent import Worker
from orangecontrib.bioinformatics.ncbi import taxonomy
from orangecontrib.bioinformatics.ncbi.gene import GeneMatcher, NCBI_ID, GENE_MATCHER_HEADER, NCBI_DETAIL_LINK

from functools import lru_cache


class GeneInfoModel(itemmodels.PyTableModel):
    def __init__(self,  *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.header_labels, self.gene_attributes = GENE_MATCHER_HEADER
        self.setHorizontalHeaderLabels(self.header_labels)

        try:
            # note: make sure ncbi_id is set in GENE_MATCHER_HEADER
            self.entrez_column_index = self.gene_attributes.index('ncbi_id')
        except ValueError as e:
            raise ValueError("Make sure 'ncbi_id' is set in gene.GENE_MATCHER_HEADER")

        self.genes = None
        self.table = None

        self.font = QFont()
        self.font.setUnderline(True)
        self.color = QColor(Qt.blue)

        @lru_cache(maxsize=10000)
        def _row_instance(row, column):
            return self[int(row)][int(column)]
        self._row_instance = _row_instance

    def initialize(self, list_of_genes):
        self.genes = list_of_genes
        self.__table_from_genes([gene for gene in list_of_genes if gene.ncbi_id])
        self.update_model()

    def rowCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else len(self._table)

    def columnCount(self, parent=QModelIndex()):
        return 0 if (parent.isValid() or self._table.size == 0) else self._table.shape[1]

    def clear(self):
        self.beginResetModel()
        self._table = np.array([[]])
        self.resetSorting()
        self._roleData.clear()
        self.endResetModel()

    def data(self, index, role,
             _str=str,
             _Qt_DisplayRole=Qt.DisplayRole,
             _Qt_EditRole=Qt.EditRole,
             _Qt_FontRole=Qt.FontRole,
             _Qt_ForegroundRole=Qt.ForegroundRole,
             _LinkRolee=LinkRole,
             _recognizedRoles=frozenset([Qt.DisplayRole, Qt.EditRole, Qt.FontRole, Qt.ForegroundRole, LinkRole])):

        if role not in _recognizedRoles:
            return None

        row, col = index.row(), index.column()
        if not 0 <= row <= self.rowCount():
            return None
        row = self.mapToSourceRows(row)

        try:
            # value = self[row][col]
            value = self._row_instance(row, col)
        except IndexError:
            return

        if role == Qt.DisplayRole:
            return QVariant(str(value))
        elif role == Qt.ToolTipRole:
            return QVariant(str(value))

        if col == self.entrez_column_index:
            if role == _Qt_ForegroundRole:
                return self.color
            elif role == _Qt_FontRole:
                return self.font
            elif role == _LinkRolee:
                return NCBI_DETAIL_LINK.format(value)

    def __table_from_genes(self, list_of_genes):
        # type: (list) -> None
        self.table = np.asarray([gene.to_list() for gene in list_of_genes])

    def get_filtered_genes(self):
        return list(self._table[:, self.entrez_column_index]) if self._table.size else []

    def filter_table(self, filter_pattern: str):
        selection = np.full(self.table.shape, True)
        for search_word in filter_pattern.split():
            match_result = np.core.defchararray.find(np.char.lower(self.table), search_word.lower()) >= 0
            selection = selection & match_result
        return selection

    def update_model(self, filter_pattern=''):
        # clear cache if model changes
        self._row_instance.cache_clear()

        if self.table.size:
            self.wrap(self.table[self.filter_table(filter_pattern).any(axis=1), :])


class UnknownGeneInfoModel(itemmodels.PyListModel):
    def __init__(self,  *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.header_labels = ['IDs from the input data without corresponding Entrez ID']
        self.genes = []

    def initialize(self, list_of_genes):
        self.genes = list_of_genes
        self.wrap([', '.join([gene.input_name for gene in list_of_genes if not gene.ncbi_id])])

    def data(self, index, role=Qt.DisplayRole):
        row = index.row()
        if role in [self.list_item_role, Qt.EditRole] and self._is_index_valid(index):
            return self[row]
        elif role == Qt.TextAlignmentRole:
            return Qt.AlignLeft | Qt.AlignTop
        elif self._is_index_valid(row):
            return self._other_data[row].get(role, None)

    def headerData(self, section, orientation, role=Qt.DisplayRole):

        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self.header_labels[section]
        return QAbstractTableModel.headerData(self, section, orientation, role)


class OWGenes(OWWidget):
    name = "Genes"
    description = "Tool for working with genes"
    icon = "../widgets/icons/OWGeneInfo.svg"
    priority = 5
    want_main_area = True

    selected_organism: int = Setting(11)
    search_pattern: str = Setting('')
    exclude_unmatched = Setting(True)
    replace_id_with_symbol = Setting(True)
    auto_commit = Setting(True)

    settingsHandler = DomainContextHandler()
    selected_gene_col = ContextSetting(None)
    use_attr_names = ContextSetting(True)

    replaces = [
        'orangecontrib.bioinformatics.widgets.OWGeneNameMatcher.OWGeneNameMatcher'
    ]

    class Inputs:
        data_table = Input("Data", Table)

    class Outputs:
        data_table = Output("Data", Table)
        gene_matcher_results = Output("Genes", Table)

    class Information(OWWidget.Information):
        pass

    def sizeHint(self):
        return QSize(1280, 960)

    def __init__(self):
        super().__init__()
        # ATTRIBUTES #
        self.target_database = NCBI_ID

        # input data
        self.input_data = None
        self.input_genes = None
        self.tax_id = None
        self.column_candidates = []

        # input options
        self.organisms = []

        # gene matcher
        self.gene_matcher = None

        # threads
        self.threadpool = QThreadPool(self)
        self.workers = None

        # progress bar
        self.progress_bar = None

        self._timer = QTimer()
        self._timer.timeout.connect(self._apply_filter)
        self._timer.setSingleShot(True)

        # GUI SECTION #

        # Control area
        self.info_box = widgetLabel(
            widgetBox(self.controlArea, "Info", addSpace=True), 'No data on input.\n'
        )

        organism_box = vBox(self.controlArea, 'Organism')
        self.organism_select_combobox = comboBox(organism_box, self,
                                                 'selected_organism',
                                                 callback=self.on_input_option_change)

        self.get_available_organisms()
        self.organism_select_combobox.setCurrentIndex(self.selected_organism)

        box = widgetBox(self.controlArea, 'Gene IDs in the input data')
        self.gene_columns_model = itemmodels.DomainModel(valid_types=(StringVariable, DiscreteVariable))
        self.gene_column_combobox = comboBox(box, self, 'selected_gene_col',
                                             label='Stored in data column',
                                             model=self.gene_columns_model,
                                             sendSelectedValue=True,
                                             callback=self.on_input_option_change)

        self.attr_names_checkbox = checkBox(box, self, 'use_attr_names', 'Stored as feature (column) names',
                                            disables=[(-1, self.gene_column_combobox)],
                                            callback=self.on_input_option_change)

        self.gene_column_combobox.setDisabled(bool(self.use_attr_names))

        output_box = vBox(self.controlArea, 'Output')

        # separator(output_box)
        # output_box.layout().addWidget(horizontal_line())
        # separator(output_box)
        self.exclude_radio = checkBox(output_box, self,
                                      'exclude_unmatched',
                                      'Exclude unmatched genes',
                                      callback=self.commit)

        self.replace_radio = checkBox(output_box, self,
                                      'replace_id_with_symbol',
                                      'Replace feature IDs with gene names',
                                      callback=self.commit)

        auto_commit(self.controlArea, self, "auto_commit", "&Commit", box=False)

        rubber(self.controlArea)

        # Main area
        self.filter = lineEdit(self.mainArea, self,
                               'search_pattern', 'Filter:',
                               callbackOnType=True, callback=self.handle_filter_callback)
        # rubber(self.radio_group)
        self.mainArea.layout().addWidget(self.filter)

        # set splitter
        self.splitter = QSplitter()
        self.splitter.setOrientation(Qt.Vertical)

        self.table_model = GeneInfoModel()
        self.table_view = QTableView()
        self.table_view.setAlternatingRowColors(True)
        self.table_view.viewport().setMouseTracking(True)
        self.table_view.setSortingEnabled(True)
        self.table_view.setShowGrid(False)
        self.table_view.verticalHeader().hide()
        # self.table_view.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)

        self.unknown_model = UnknownGeneInfoModel()

        self.unknown_view = QTableView()
        self.unknown_view.setModel(self.unknown_model)
        self.unknown_view.verticalHeader().hide()
        self.unknown_view.setShowGrid(False)
        self.unknown_view.setSelectionMode(QAbstractItemView.NoSelection)
        self.unknown_view.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)

        self.splitter.addWidget(self.table_view)
        self.splitter.addWidget(self.unknown_view)

        self.splitter.setStretchFactor(0, 90)
        self.splitter.setStretchFactor(1, 10)

        self.mainArea.layout().addWidget(self.splitter)

    def handle_filter_callback(self):
        self._timer.stop()
        self._timer.start(500)

    def _apply_filter(self):
        # filter only if input data is present and model is populated
        if self.table_model.table is not None:
            self.table_model.update_model(filter_pattern=str(self.search_pattern))
            self.commit()

    def __reset_widget_state(self):
        self.table_view.clearSpans()
        self.table_view.setModel(None)
        self.table_model.clear()
        self.unknown_model.clear()
        self._update_info_box()

    def _update_info_box(self):

        if self.input_genes and self.gene_matcher:
            num_genes = len(self.gene_matcher.genes)
            known_genes = len(self.gene_matcher.get_known_genes())

            info_text = '{} genes in input data\n' \
                        '{} genes match Entrez database\n' \
                        '{} genes with match conflicts\n'.format(num_genes, known_genes, num_genes - known_genes)

        else:
            info_text = 'No data on input.'

        self.info_box.setText(info_text)

    def _progress_advance(self):
        # GUI should be updated in main thread. That's why we are calling advance method here
        if self.progress_bar:
            self.progress_bar.advance()

    def _handle_matcher_results(self):
        assert threading.current_thread() == threading.main_thread()

        if self.progress_bar:
            self.progress_bar.finish()
            self.setStatusMessage('')

        # update info box
        self._update_info_box()

        # set output options
        self.toggle_radio_options()

        # set known genes
        self.table_model.initialize(self.gene_matcher.genes)
        self.table_view.setModel(self.table_model)
        self.table_view.selectionModel().selectionChanged.connect(self.commit)
        self.table_view.setSelectionBehavior(QAbstractItemView.SelectRows)

        self.table_view.setItemDelegateForColumn(
            self.table_model.entrez_column_index, LinkStyledItemDelegate(self.table_view)
        )
        v_header = self.table_view.verticalHeader()
        option = self.table_view.viewOptions()
        size = self.table_view.style().sizeFromContents(
            QStyle.CT_ItemViewItem, option,
            QSize(20, 20), self.table_view)

        v_header.setDefaultSectionSize(size.height() + 2)
        v_header.setMinimumSectionSize(5)
        self.table_view.horizontalHeader().setStretchLastSection(True)

        # set unknown genes
        self.unknown_model.initialize(self.gene_matcher.genes)
        self.unknown_view.verticalHeader().setStretchLastSection(True)

        self._apply_filter()

    def get_available_organisms(self):
        available_organism = sorted([(tax_id, taxonomy.name(tax_id)) for tax_id in taxonomy.common_taxids()],
                                    key=lambda x: x[1])

        self.organisms = [tax_id[0] for tax_id in available_organism]
        self.organism_select_combobox.addItems([tax_id[1] for tax_id in available_organism])

    def gene_names_from_table(self):
        """ Extract and return gene names from `Orange.data.Table`.
        """
        self.input_genes = []
        if self.input_data:
            if self.use_attr_names:
                self.input_genes = [str(attr.name).strip() for attr in self.input_data.domain.attributes]
            else:
                if self.selected_gene_col is None:
                    self.selected_gene_col = self.gene_column_identifier()

                self.input_genes = [str(e[self.selected_gene_col]) for e in self.input_data
                                    if not np.isnan(e[self.selected_gene_col])]

    def _update_gene_matcher(self):
        self.gene_names_from_table()
        self.gene_matcher = GeneMatcher(self.get_selected_organism(), case_insensitive=True)
        self.gene_matcher.genes = self.input_genes
        self.gene_matcher.organism = self.get_selected_organism()

    def get_selected_organism(self):
        return self.organisms[self.selected_organism]

    def match_genes(self):
        if self.gene_matcher:
            # init progress bar
            self.progress_bar = ProgressBar(self, iterations=len(self.gene_matcher.genes))
            # status message
            self.setStatusMessage('Gene matcher running')

            worker = Worker(self.gene_matcher.run_matcher, progress_callback=True)
            worker.signals.progress.connect(self._progress_advance)
            worker.signals.finished.connect(self._handle_matcher_results)

            # move download process to worker thread
            self.threadpool.start(worker)

    def on_input_option_change(self):
        self.__reset_widget_state()
        self._update_gene_matcher()
        self.match_genes()

    def gene_column_identifier(self):
        """
        Get most suitable column that stores genes. If there are
        several suitable columns, select the one with most unique
        values. Take the best one.
        """

        # candidates -> (variable, num of unique values)
        candidates = ((col, np.unique(self.input_data.get_column_view(col)[0]).size)
                      for col in self.gene_columns_model
                      if isinstance(col, DiscreteVariable) or isinstance(col, StringVariable))

        best_candidate, _ = sorted(candidates, key=lambda x: x[1])[-1]
        return best_candidate

    def find_genes_location(self):
        """ Try locate the genes in the input data when we first load the data.

            Proposed rules:
                - when no suitable feature names are present, check the columns.
                - find the most suitable column, that is, the one with most unique values.

        """
        domain = self.input_data.domain
        if not domain.attributes:
            if self.selected_gene_col is None:
                self.selected_gene_col = self.gene_column_identifier()
                self.use_attr_names = False

    @Inputs.data_table
    def handle_input(self, data):
        self.closeContext()
        self.input_data = None
        self.input_genes = None
        self.__reset_widget_state()
        self.gene_columns_model.set_domain(None)
        self.selected_gene_col = None

        if data:
            self.input_data = data
            self.gene_columns_model.set_domain(self.input_data.domain)

            # check if input table has tax_id, human is used if tax_id is not found
            self.tax_id = str(self.input_data.attributes.get(TAX_ID, '9606'))
            # check for gene location. Default is that genes are attributes in the input table.
            self.use_attr_names = self.input_data.attributes.get(GENE_AS_ATTRIBUTE_NAME, self.use_attr_names)

            if self.tax_id in self.organisms and not self.selected_organism:
                self.selected_organism = self.organisms.index(self.tax_id)

            self.openContext(self.input_data.domain)
            self.find_genes_location()
            self.on_input_option_change()

    def commit(self):
        selection = self.table_view.selectionModel().selectedRows(self.table_model.entrez_column_index)

        selected_genes = [row.data() for row in selection]
        if not len(selected_genes):
            selected_genes = self.table_model.get_filtered_genes()

        gene_ids = self.get_target_ids()
        known_genes = [gid for gid in gene_ids if gid != '?']

        table = None
        gm_table = None
        if known_genes:
            # Genes are in rows (we have a column with genes).
            if not self.use_attr_names:

                if self.target_database in self.input_data.domain:
                    gene_var = self.input_data.domain[self.target_database]
                    metas = self.input_data.domain.metas
                else:
                    gene_var = StringVariable(self.target_database)
                    metas = self.input_data.domain.metas + (gene_var,)

                domain = Domain(self.input_data.domain.attributes,
                                self.input_data.domain.class_vars,
                                metas)

                table = self.input_data.transform(domain)
                col, _ = table.get_column_view(gene_var)
                col[:] = gene_ids

                # filter selected rows
                selected_rows = [row_index for row_index, row in enumerate(table)
                                 if str(row[gene_var]) in selected_genes]

                # handle table attributes
                table.attributes[TAX_ID] = self.get_selected_organism()
                table.attributes[GENE_AS_ATTRIBUTE_NAME] = False
                table.attributes[GENE_ID_COLUMN] = self.target_database
                table = table[selected_rows] if selected_rows else table

                if self.exclude_unmatched:
                    # create filter from selected column for genes
                    only_known = table_filter.FilterStringList(gene_var, known_genes)
                    # apply filter to the data
                    table = table_filter.Values([only_known])(table)

                self.Outputs.data_table.send(table)

            # genes are are in columns (genes are features).
            else:
                domain = self.input_data.domain.copy()
                table = self.input_data.transform(domain)

                for gene in self.gene_matcher.genes:
                    if gene.input_name in table.domain:

                        table.domain[gene.input_name].attributes[self.target_database] = \
                            str(gene.ncbi_id) if gene.ncbi_id else '?'

                        if self.replace_id_with_symbol:
                            try:
                                table.domain[gene.input_name].name = str(gene.symbol)
                            except AttributeError:
                                # TODO: missing gene symbol, need to handle this?
                                pass

                # filter selected columns
                selected = [column for column in table.domain.attributes
                            if self.target_database in column.attributes and
                            str(column.attributes[self.target_database]) in selected_genes]

                output_attrs = table.domain.attributes

                if selected:
                    output_attrs = selected

                if self.exclude_unmatched:
                    output_attrs = [col for col in output_attrs if col.attributes[self.target_database] in known_genes]

                domain = Domain(output_attrs,
                                table.domain.class_vars,
                                table.domain.metas)

                table = table.from_table(domain, table)

                # handle table attributes
                table.attributes[TAX_ID] = self.get_selected_organism()
                table.attributes[GENE_AS_ATTRIBUTE_NAME] = True
                table.attributes[GENE_ID_ATTRIBUTE] = self.target_database

            gm_table = self.gene_matcher.to_data_table(selected_genes=selected_genes if selected_genes else None)

        self.Outputs.data_table.send(table)
        self.Outputs.gene_matcher_results.send(gm_table)

    def toggle_radio_options(self):
        self.replace_radio.setEnabled(bool(self.use_attr_names))

        if self.gene_matcher.genes:
            # enable checkbox if unknown genes are detected
            self.exclude_radio.setEnabled(len(self.gene_matcher.genes) != len(self.gene_matcher.get_known_genes()))
            self.exclude_unmatched = len(self.gene_matcher.genes) != len(self.gene_matcher.get_known_genes())

    def get_target_ids(self):
        return [str(gene.ncbi_id) if gene.ncbi_id else '?' for gene in self.gene_matcher.genes]


if __name__ == "__main__":
    def main_test():
        from AnyQt.QtWidgets import QApplication

        app = QApplication([])
        w = OWGenes()
        w.show()
        w.raise_()
        r = app.exec_()
        w.saveSettings()
        return r

    sys.exit(main_test())



""" Resolwe module """
from .genapi import GenAPI, DEFAULT_EMAIL, DEFAULT_PASSWD, cache_backend, cache_name


def connect(username, password, url, server_type):
    """ Connect to Resolwe server

    :param username:
    :type username: str

    :param password:
    :type password: str

    :param url:
    :type url: str

    :param server_type: genesis or resolwe
    :type server_type: str

    :return: Instance of GenAPI or ResolweAPI
    """

    if server_type == 'genesis':
        try:
            return GenAPI(username, password, url)
        except Exception as e:
            print(e)
            raise ResolweAuthException(e.args[0]) from e
    else:
        """ Not yet supported """
        pass


class ResolweAuthException(Exception):
    """ A login error occurred. """



import unittest

import Orange
import numpy as np
from Orange.clustering import DBSCAN, KMeans
from Orange.data import Domain, ContinuousVariable, Table, DiscreteVariable

from orangecontrib.bioinformatics.annotation.annotate_projection import \
    annotate_projection, labels_locations, get_epsilon, compute_concave_hulls, \
    assign_labels


class TestAnnotateProjection(unittest.TestCase):

    def setUp(self):
        domain = Domain([ContinuousVariable("x"), ContinuousVariable("y")])
        data = np.array([[1., 1.], [1., 1.5], [1.5, 1.], [1.2, 1.2], [1.7, 1.8],
                         [10., 10.], [10., 10.5], [10.5, 11.], [10.5, 10.7],
                         [10.5, 10.6], [10.5, 10.7], [10.5, 10.7]])
        self.data = Table(domain, data)

        domain_ann = Domain([ContinuousVariable("a"), ContinuousVariable("b"),
                            ContinuousVariable("c")])
        data_ann = np.array([[0.9, 1, 0.4], [0.9, 0.4, 0.2], [0.9, 0.4, 0.2],
                             [0.9, 1, 0.2], [0.9, 0.4, 0.2],
                             [0.2, 0.4, 0.7], [0.1, 0.4, 0.6], [0.2, 0.9, 0.2],
                             [0.1, 0.4, 0.6], [0.1, 0.8, 0.6], [0.1, 0.8, 0.6],
                             [0.1, 0.8, 0.6]])
        self.annotations = Table(domain_ann, data_ann)

    def test_annotate_projection(self):
        clusters, clusters_meta, eps = annotate_projection(
            self.annotations, self.data, clustering_algorithm=DBSCAN,
            labels_per_cluster=2, eps=3)

        self.assertLessEqual(int(Orange.__version__.split(".")[1]), 22)
        # TODO: uncomment when new Orange released, and remove version check
        # self.assertListEqual(
        #     list(map(clusters.domain.attributes[0].repr_val, clusters.X[:, 0])),
        #     ["0", "0", "0", "0", "0", "1", "1", "1", "1", "1"])

        # self.assertEqual(len(clusters_meta), 2)
        # self.assertEqual(len(clusters_meta["C0"]), 3)
        # self.assertEqual(len(clusters_meta["C1"]), 3)
        # self.assertEqual(clusters_meta["C1"][0][0][0], 'c')
        # self.assertAlmostEqual(clusters_meta["C1"][0][0][1], 0.6, 5)
        # self.assertEqual(clusters_meta["C1"][0][1][0], 'b')
        # self.assertAlmostEqual(clusters_meta["C1"][0][1][1], 0.4, 5)

        # self.assertEqual(2, len(clusters_meta["C1"][1]))
        # self.assertEqual(2, clusters_meta["C1"][2].shape[1])
        # self.assertEqual(
        #     type(np.ndarray), type(clusters_meta["C1"][2].shape[1]))
        # self.assertEqual(float, type(eps))

    def test_example_not_clustered(self):
        self.data[-1] = [23, 23]
        clusters, clusters_meta, eps = annotate_projection(
            self.annotations, self.data, clustering_algorithm=DBSCAN,
            eps=2, min_samples=3)

        self.assertLessEqual(int(Orange.__version__.split(".")[1]), 22)
        # TODO: uncomment when new Orange released, and remove version check
        # self.assertListEqual(
        #     list(map(clusters.domain.attributes[0].repr_val, clusters.X[:, 0])),
        #     ["0", "0", "0", "0", "0", "1", "1", "1", "1", "-1"])

        # self.assertEqual(len(clusters_meta), 2)
        # self.assertEqual(len(clusters_meta["C0"]), 3)
        # self.assertEqual(len(clusters_meta["C1"]), 3)
        # self.assertEqual(clusters_meta["C1"][0][0][0], 'c')
        # self.assertAlmostEqual(clusters_meta["C1"][0][0][1], 0.75, 5)
        # self.assertEqual(clusters_meta["C1"][0][1][0], 'b')
        # self.assertAlmostEqual(clusters_meta["C1"][0][1][1], 0.25, 5)

    def test_other_clustering(self):
        clusters, clusters_meta, locs = annotate_projection(
            self.annotations, self.data,
            clustering_algorithm=KMeans, n_clusters=2,
            labels_per_cluster=2)

        self.assertEqual(2, len(set(clusters.X[:, 0].flatten())))
        self.assertEqual(3, len(clusters_meta["C1"]))
        self.assertEqual(3, len(clusters_meta["C2"]))
        self.assertEqual(2, len(clusters_meta["C1"][0]))
        self.assertEqual(2, len(clusters_meta["C2"][0]))

    def test_labels_location(self):
        clusters = Table(Domain([DiscreteVariable("cl", values=["1", "2"])]),
                         [[0]] * 6 + [[1]] * 6)
        locs = labels_locations(self.data, clusters)

        self.assertEqual(dict, type(locs))
        self.assertEqual(tuple, type(locs["1"]))
        self.assertEqual(tuple, type(locs["2"]))

    def test_get_epsilon(self):
        data = Table("iris")
        eps = get_epsilon(data)

        self.assertGreaterEqual(eps, 0.9)

    def test_compute_concave_hulls(self):
        data = Table("iris")[:, 2:4]
        clusters = Table(
            Domain([DiscreteVariable("cl", values=["1", "2", "3"])]),
            [[0]] * 50 + [[1]] * 50 + [[2]] * 50
        )
        hulls = compute_concave_hulls(data, clusters, epsilon=0.5)

        self.assertEqual(3, len(hulls))
        self.assertEqual(2, hulls["1"].shape[1])  # hull have x and y
        self.assertEqual(2, hulls["2"].shape[1])  # hull have x and y
        self.assertEqual(2, hulls["3"].shape[1])  # hull have x and y

    def test_empty_annotations(self):
        ann = Table(Domain([]), np.empty((len(self.data), 0)))
        clusters, clusters_meta, eps = annotate_projection(ann, self.data)

        self.assertLessEqual(int(Orange.__version__.split(".")[1]), 22)
        # TODO: uncomment when new Orange released, and remove version check
        # self.assertGreater(len(clusters), 0)
        # self.assertGreater(len(clusters_meta), 0)
        # self.assertEqual(0, len(clusters_meta["C1"][0]))
        # self.assertGreater(len(clusters_meta["C1"][1]), 0)
        # self.assertGreater(len(clusters_meta["C1"][2]), 0)

    def test_one_label(self):
        """
        Test whether having only one label works fine, in this case one cluster
        will not have a label assigned - this must work fine as well.
        """
        domain_ann = Domain([ContinuousVariable("a")])
        data_ann = np.array([[0.9], [0.9], [0.9], [0.9], [0.9],
                             [0], [0], [0], [0], [0], [0], [0]])
        ann = Table(domain_ann, data_ann)
        clusters, clusters_meta, eps = annotate_projection(ann, self.data,
                                                           eps=2)
        self.assertGreater(len(clusters), 0)
        self.assertGreater(len(clusters_meta), 0)
        self.assertGreater(len(clusters_meta["C1"][0]), 0)
        self.assertGreater(len(clusters_meta["C1"][1]), 0)
        self.assertGreater(len(clusters_meta["C1"][2]), 0)

    def test_socres_with_nan(self):
        self.annotations.X[0, 0] = np.nan
        self.annotations.X[1, 1] = np.nan
        self.annotations.X[1, 2] = np.nan

        self.annotations.X[2, 0] = np.nan
        self.annotations.X[2, 1] = np.nan
        self.annotations.X[2, 2] = np.nan

        clusters = Table(
            Domain([DiscreteVariable("Cluster", values=["C1", "C3"])]),
            np.array([[0] * 5 + [1] * 7]).reshape(-1, 1))

        labels_dict, labels_items = assign_labels(clusters, self.annotations, 3)

        transformed_labels = list(
            map(labels_items.domain.attributes[0].repr_val,
                labels_items.X[:, 0]))
        self.assertListEqual(
            ['b', 'a', '?', 'b', 'a', 'c', 'c', 'b', 'c', 'b', 'b', 'b'],
            transformed_labels
        )
        self.assertTupleEqual((0.4, 0.4), list(zip(*labels_dict["C1"]))[1])

        # check all nans
        self.annotations.X[:, :] = np.nan
        labels_dict, labels_items = assign_labels(clusters, self.annotations, 3)
        transformed_labels = list(
            map(labels_items.domain.attributes[0].repr_val,
                labels_items.X[:, 0]))
        self.assertListEqual(
            ['?'] * len(self.annotations),
            transformed_labels
        )
        self.assertEqual(0, len(labels_dict["C1"]))
        self.assertEqual(0, len(labels_dict["C3"]))



DOMAIN = "dictybase"
# FILENAME = "information_mappings.pck"
# TAGS = ["Dictyostelium discoideum", "gene", "essential", "dictyBase"]

PHENOTYPES_FILENAME = "mutants.pkl"
PHENOTYPES_TAGS = ["Dictyostelium discoideum", "mutant", "dictyBase", "phenotype"]
PHENOTYPES_TITLE = 'dictyBase mutant phenotypes'

DICTY_INFO_URL = 'http://www.dictybase.org/db/cgi-bin/dictyBase/download/download.pl?area=general&ID=gene_information.txt'
DICTY_MAPPING_URL = 'http://www.dictybase.org/db/cgi-bin/dictyBase/download/download.pl?area=general&ID=DDB-GeneID-UniProt.txt'


from six import StringIO
import doctest

import unittest

from orangecontrib.bio.kegg.entry import parser, fields, DBEntry, entry_decorate


TEST_ENTRY = """\
ENTRY       test_id    something else
NAME        test
DESCRIPTION This is a test's description.
            It spans
            multiple lines
  SUB       This is a description's sub
            section
///
"""


@entry_decorate
class Entry(DBEntry):
    pass


class TestEntry(unittest.TestCase):
    def test_entry(self):
        """
        Test basic DBEntry class.
        """
        entry = Entry(TEST_ENTRY)
        self.assertEqual(entry.entry_key, "test_id")
        self.assertEqual(entry.ENTRY.TITLE, "ENTRY")
        self.assertEqual(str(entry), TEST_ENTRY[:-4])


class TestParser(unittest.TestCase):
    def test_parser(self):
        parse = parser.DBGETEntryParser()
        stream = StringIO(TEST_ENTRY)

        expected = [
            (parse.ENTRY_START, None, None),
            (parse.SECTION_START, "ENTRY", "test_id    something else\n"),
            (parse.SECTION_END, "ENTRY", None),
            (parse.SECTION_START, "NAME", "test\n"),
            (parse.SECTION_END, "NAME", None),
            (parse.SECTION_START, "DESCRIPTION", "This is a test's description.\n"),
            (parse.TEXT, None, "It spans\n"),
            (parse.TEXT, None, "multiple lines\n"),
            (parse.SUBSECTION_START, "SUB", "This is a description's sub\n"),
            (parse.TEXT, None, "section\n"),
            (parse.SUBSECTION_END, "SUB", None),
            (parse.SECTION_END, "DESCRIPTION", None),
            (parse.ENTRY_END, None, None)
        ]
        self.assertSequenceEqual(list(parse.parse(stream)), expected)


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(parser))
    return tests



""" Orange data table and other data manipulation utilities """
import numbers
import numpy
import Orange.data

from operator import itemgetter
from typing import Sequence, Tuple

from Orange.widgets.widget import OWWidget, Msg

# Data hints variables

# species
TAX_ID = 'taxonomy_id'

# Will be set to True if gene names are represented as attribute names.
# If gene names are in rows, we set this value to False. (user must select proper column index)
GENE_AS_ATTRIBUTE_NAME = 'gene_as_attribute_name'

# Name of the column where rows are gene ids
GENE_ID_COLUMN = 'gene_id_column'

# Name of the variable attribute that holds gene id
GENE_ID_ATTRIBUTE = 'gene_id_attribute'

# Error strings
ERROR_ON_MISSING_ANNOTATION = 'Missing annotation on gene IDs and organism in the input data.'
ERROR_ON_MISSING_GENE_ID = 'Missing gene ID information in the input data'
ERROR_ON_MISSING_TAX_ID = 'Missing organism information in the input data'

col_spec = Sequence[Tuple[Orange.data.Variable, Sequence[numbers.Real]]]
metas_spec = Sequence[Tuple[Orange.data.Variable, Sequence[str]]]


def append_columns(data, attributes=(), class_vars=(), metas=()):
    # type: (Orange.data.Table, col_spec, col_spec, metas_spec) -> Orange.data.Table
    """ Append a set of columns to a data table.


    Parameters
    ----------
    data : Orange.data.Table
        Primary table.
    attributes : Sequence[Tuple[Orange.data.Variable], Sequence[float]]
        A Sequence of variable and column data tuples to append to the
        `data`.
    class_vars : Sequence[Tuple[Orange.data.Variable], Sequence[float]]
        A Sequence of variable and column data tuples to append to the
        `data` in the
    metas : Sequence[Tuple[Orange.data.Variable], Sequence[str]]
        A Sequence of variable and column data tuples to append to the
        `data`

    Returns
    -------
    data : Orange.data.Table
        A copy of the original `data` input extended with all columns from
        `attributes`, `class_vars`, `metas` parameters

    Note
    ----
    All variables in the original and new columns should be distinct.
    """
    domain = data.domain
    new_attributes = tuple(map(itemgetter(0), attributes))
    new_class_vars = tuple(map(itemgetter(0), class_vars))
    new_metas = tuple(map(itemgetter(0), metas))

    new_domain = Orange.data.Domain(
        domain.attributes + new_attributes,
        domain.class_vars + new_class_vars,
        domain.metas + new_metas
    )

    def ascolumn(array, n):
        # type: (Sequence[float], int) -> numpy.ndarray
        array = numpy.asarray(array)
        if array.ndim < 2:
            array = array.reshape((n, 1))
        return array
    N = len(data)

    attr_cols = [ascolumn(col, N) for _, col in attributes]
    class_cols = [ascolumn(col, N) for _, col in class_vars]
    meta_cols = [ascolumn(col, N) for _, col in metas]

    new_data = data.from_table(new_domain, data)

    for i, (var, col) in enumerate(zip(new_attributes, attr_cols),
                                   start=len(domain.attributes)):
        assert new_data.domain.attributes[i] is var
        new_data.X[:, i] = col.ravel()

    for i, (var, col) in enumerate(zip(new_class_vars, class_cols),
                                   start=len(domain.class_vars)):
        assert new_data.domain.class_vars[i] is var
        new_data._Y[:, i] = col.ravel()

    for i, (var, col) in enumerate(zip(new_metas, meta_cols),
                                   start=len(domain.metas)):
        assert new_data.domain.metas[i] is var
        new_data.metas[:, i] = col.ravel()

    return new_data



"""
This module cluster the projection of data (usually it is 2D projection) with
one of the standard algorithms and attach a certain number of labels per
cluster.

Example:

>>> from Orange.projection import TSNE
>>> from Orange.data import Table
>>> from orangecontrib.bioinformatics.utils import serverfiles
>>> from orangecontrib.bioinformatics.annotation.annotate_projection import \
...     annotate_projection
>>> from orangecontrib.bioinformatics.annotation.annotate_samples import \
...     AnnotateSamples
>>>
>>> # load data
>>> data = Table("https://datasets.orange.biolab.si/sc/aml-1k.tab.gz")
>>> marker_p = serverfiles.localpath_download(
...     'marker_genes','panglao_gene_markers.tab')
>>> markers = Table(marker_p)
>>>
>>> # annotate data with labels
>>> annotator = AnnotateSamples()
>>> annotations = annotator.annotate_samples(data, markers)
>>>
>>> # project data in 2D
>>> tsne = TSNE(n_components=2)
>>> tsne_model = tsne(data)
>>> embedding = tsne_model(data)
>>>
>>> # get clusters and annotations for clusters
>>> clusters, clusters_meta, eps = annotate_projection(annotations, embedding,
...     clustering_algorithm=DBSCAN, eps=1.2)

In case when user uses a DBSCAN algorithm and do not provide eps to the
`annotate_projection` function it is computed automatically with knn method.

>>> clusters, clusters_meta, eps = annotate_projection(annotations, embedding,
...     clustering_algorithm=DBSCAN)

"""


from collections import Counter

from Orange.clustering import DBSCAN
import numpy as np
from Orange.data import Domain, DiscreteVariable, Table
from scipy.spatial import distance
import shapely.geometry as geometry
from scipy.spatial import Delaunay
from shapely.ops import cascaded_union, polygonize
from math import sqrt


def cluster_data(coordinates, clustering_algorithm=DBSCAN, **kwargs):
    """
    This function receives data and cluster them.

    Parameters
    ----------
    coordinates : Orange.data.Table
        Visualisation coordinates - embeddings
    clustering_algorithm : callable
        Algorithm used for clustering.

    Returns
    -------
    Orange.data.Table
        List of cluster indices.
    """
    learner = clustering_algorithm(**kwargs)
    clusters = learner(coordinates)
    if not isinstance(clusters, np.ndarray):  # old clustering method
        clusters = clusters(coordinates)
        clusters = np.array(list(map(int, map(
            clusters.domain.attributes[0].repr_val, clusters.X[:, 0]
        )))).flatten()

    # sort classes in descending order base on number of cases in the cluster
    sorted_clust_idx = [
        v for v, _ in Counter(clusters).most_common() if v != -1]

    # re-indexed array
    new_clustering = np.empty(len(clusters))
    new_clustering[:] = np.nan  # nan for not clustered
    # reindex based on descending cluster size
    for i, v in enumerate(sorted_clust_idx):
        new_clustering[clusters == v] = i

    # create the table
    new_domain = Domain([DiscreteVariable(
        "Clusters", values=[
            "C{}".format(i) for i in range(1, len(sorted_clust_idx) + 1)])])
    return Table(new_domain, new_clustering.reshape((-1, 1)))


def assign_labels(clusters, annotations, labels_per_cluster):
    """
    This function assigns a certain number of labels per cluster. Each cluster
    gets `labels_per_cluster` number of most common labels in cluster assigned.

    Parameters
    ----------
    clusters : Orange.data.Table
        Cluster indices for each item.
    annotations : Orange.data.Table
        Table with annotations and their probabilities.
    labels_per_cluster : int
        Number of labels that need to be assigned to each cluster.

    Returns
    -------
    dict
        Dictionary with cluster index as a key and list of annotations as a
        value. Each list include tuples with the annotation name and their
        proportion in the cluster.
    Orange.data.Table
        The array with the annotation assigned to the item.
    """
    clusters_unique = set(clusters.domain[0].values)

    if len(annotations.domain) == 0:
        return {cl: [] for cl in clusters_unique}, \
               Table(Domain([DiscreteVariable("Annotation", values=[])]),
                     np.ones((len(clusters), 1)) * np.nan)

    labels = np.array(list(map(str, annotations.domain.attributes)))

    # remove rows with all nans
    nan_mask = np.isnan(annotations.X).all(axis=1)
    ann_not_nan = annotations.X[~nan_mask]

    # find indices and labels
    annotation_best_idx = np.nanargmax(ann_not_nan, axis=1)
    annotation_best = labels[annotation_best_idx]

    # join back together
    items_annotations = np.empty(annotations.X.shape[0], dtype=labels.dtype)
    items_annotations[~nan_mask] = annotation_best

    annotations_clusters = {}
    for cl in clusters_unique:
        mask = np.array(list(
            map(clusters.domain.attributes[0].repr_val,
                clusters.X[:, 0]))).flatten() == cl
        labels_cl = items_annotations[mask]
        # remove nans from labels
        labels_cl_filtered = labels_cl[~(labels_cl == "")]

        counts = Counter(labels_cl_filtered)
        annotations_clusters[cl] = [
            (l, c / len(labels_cl))
            for l, c in counts.most_common(labels_per_cluster)]

    # pack item annotations to Table
    nan_mask = items_annotations == ""
    values, indices = np.unique(
        items_annotations[~nan_mask], return_inverse=True)
    corrected_idx = np.ones(items_annotations.shape) * np.nan
    corrected_idx[~nan_mask] = indices
    domain = Domain([DiscreteVariable("Annotation", values=values)])
    item_annotations = Table(domain, corrected_idx.reshape((-1, 1)))

    return annotations_clusters, item_annotations


def labels_locations(coordinates, clusters):
    """
    Function computes the location of the label for each cluster.
    The location is compute as a center point.

    Parameters
    ----------
    coordinates : Orange.data.Table
        Visualisation coordinates - embeddings
    clusters : Orange.data.Table
        Cluster indices for each item.

    Returns
    -------
    dict
        The coordinates for locating the label. Dictionary with cluster index
        as a key and tuple (x, y) as a value.
    """
    clusters_unique = set(
        clusters.domain[0].values) - {"-1"}  # -1 is not clustered
    locations = {}
    for cl in clusters_unique:
        mask = np.array(list(
            map(clusters.domain.attributes[0].repr_val,
                clusters.X[:, 0]))).flatten() == cl
        cl_coordinates = coordinates.X[mask, :]
        x, y = np.mean(cl_coordinates, axis=0)
        locations[cl] = (x, y)
    return locations


def get_epsilon(coordinates, k=10, skip=0.1):
    """
    The function computes the epsilon parameter for DBSCAN through method
    proposed in the paper.

    Parameters
    ----------
    coordinates : Orange.data.Table
        Visualisation coordinates - embeddings
    k : int
        Number kth observed neighbour
    skip : float
        Percentage of skipped neighborus.

    Returns
    -------
    float
        Epsilon parameter for DBSCAN
    """
    x = coordinates.X
    if len(x) > 1000:  # subsampling is required
        i = len(x) // 1000
        x = x[::i]

    d = distance.squareform(distance.pdist(x))
    kth_point = np.argpartition(d, k+1, axis=1)[:, k+1]
    # k+1 since first one is item itself
    kth_dist = np.sort(d[np.arange(0, len(kth_point)), kth_point])

    # currently mark proportion equal to skip as a noise
    return kth_dist[-int(np.round(len(kth_dist) * skip))]


def compute_concave_hulls(coordinates, clusters, epsilon):
    """
    Function computes the points of the concave hull around points.

    Parameters
    ----------
    coordinates : Orange.data.Table
        Visualisation coordinates - embeddings
    clusters : Orange.data.Table
       Cluster indices for each item.
    epsilon : float
        Epsilon used by DBSCAN to cluster the data

    Returns
    -------
    dict
       The points of the concave hull. Dictionary with cluster index
       as a key and np.ndaray of points as a value -
       [[x1, y1], [x2, y2], [x3, y3], ...]
    """

    def get_shape(points, epsilon):
        """
        Compute the shape (concave hull) of a set of a cluster.
        """
        if len(points) < 4:
            # When you have a triangle, there is no sense in computing the hull
            return geometry.MultiPoint(list(points)).convex_hull

        def add_edge(edges, edge_points, coords, i, j):
            """
            Add a line between the i-th and j-th points,
            if not in the list already
            """
            if (i, j) in edges or (j, i) in edges:
                # already added
                return
            edges.add((i, j))
            edge_points.append(coords[[i, j]])

        tri = Delaunay(points)
        edges = set()
        edge_points = []
        # loop over triangles:
        # ia, ib, ic = indices of corner points of the triangle
        for ia, ib, ic in tri.vertices:
            pa = points[ia]
            pb = points[ib]
            pc = points[ic]

            # Lengths of sides of triangle
            a = sqrt((pa[0] - pb[0]) ** 2 + (pa[1] - pb[1]) ** 2)
            b = sqrt((pb[0] - pc[0]) ** 2 + (pb[1] - pc[1]) ** 2)
            c = sqrt((pc[0] - pa[0]) ** 2 + (pc[1] - pa[1]) ** 2)

            # filter - longest edge of triangle smaller than epsilon
            if max(a, b, c) <= epsilon:
                add_edge(edges, edge_points, points, ia, ib)
                add_edge(edges, edge_points, points, ib, ic)
                add_edge(edges, edge_points, points, ic, ia)

        m = geometry.MultiLineString(edge_points)
        triangles = list(polygonize(m))
        return cascaded_union(triangles)

    hulls = {}
    clusters_array = np.array(list(map(
        clusters.domain.attributes[0].repr_val, clusters.X[:, 0])))
    for cl in set(clusters_array) - {"None", "?"}:
        points = coordinates.X[clusters_array == cl]

        # subsample when more than 1000 points
        # it keeps time finding hull under 0.3 s on my computer
        if points.shape[0] > 1000:
            points = points[np.random.randint(points.shape[0], size=1000), :]

        # epsilon * 2 seems to be good parameter for lines to be smooth enough
        concave_hull = get_shape(points, epsilon=epsilon * 2)
        # expand_and_smooth the curve - selecting epsilon for the distance
        # shows approximately what is DBSCAN neighbourhood
        # this move the hull further away and then move it back to make it more
        # smooth
        concave_hull = concave_hull.buffer(3 * epsilon).buffer(-epsilon * 2)

        hulls[cl] = np.array(list(map(list, concave_hull.exterior.coords.xy))).T

    return hulls


def annotate_projection(annotations, coordinates,
                        clustering_algorithm=DBSCAN,
                        labels_per_cluster=3, **kwargs):
    """
    Function cluster the data based on coordinates, and assigns a certain number
    of labels per cluster. Each cluster gets `labels_per_cluster` number of most
    common labels in cluster assigned.

    Parameters
    ----------
    annotations : Orange.data.Table
        Table with annotations and their probabilities.
    coordinates : Orange.data.Table
        Visualisation coordinates - embeddings
    clustering_algorithm : callable, optional (default = DBSCAN)
        Algorithm used in clustering.
    labels_per_cluster : int, optional (default = 3)
        Number of labels that need to be assigned to each cluster.

    Returns
    -------
    Orange.data.Table
        List of cluster indices.
    dict
        Dictionary with cluster index as a key and list of annotations as a
        value. Each list include tuples with the annotation name and their
        proportion in the cluster.
    dict
        The coordinates for locating the label. Dictionary with cluster index
        as a key and tuple (x, y) as a value.
    """
    assert len(annotations) == len(coordinates), \
        "Number of coordinates does not match to number of annotations"
    # sklearn clustering want to have one example
    assert len(coordinates) > 0, "At least one data point need to be provided"
    assert len(coordinates.domain) > 0, \
        "Coordinates need to have at least one attribute"

    eps = kwargs.get("eps", get_epsilon(coordinates))
    if clustering_algorithm == DBSCAN:
        kwargs["eps"] = eps

    # get clusters
    clusters = cluster_data(coordinates, clustering_algorithm, **kwargs)

    # assign top n labels to group
    annotations_cl, item_annotations = assign_labels(
        clusters, annotations, labels_per_cluster)

    labels_loc = labels_locations(coordinates, clusters)

    concave_hull = compute_concave_hulls(coordinates, clusters, eps)

    # crate the dictionary with annotations, labels locations, and hulls for
    # each cluster
    clusters_meta = {}
    for cl in annotations_cl.keys():
        clusters_meta[cl] = (
            annotations_cl[cl], labels_loc[cl], concave_hull[cl])

    # add the labels to the cluster table
    clusters_ann = Table(
        Domain(clusters.domain.attributes + item_annotations.domain.attributes),
        np.concatenate((clusters.X, item_annotations.X), axis=1))

    return clusters_ann, clusters_meta, eps



import math
import scipy
import threading
import numpy as np

from typing import Tuple, Union

from scipy.stats import hypergeom

ALT_TWO = "two-sided"
ALT_LESS = "less"
ALT_GREATER = "greater"
ALTERNATIVES = [ALT_GREATER, ALT_TWO, ALT_LESS]


def score_t_test(a, b, axis=0, alternative=ALT_TWO):
    # type: (np.array, np.array, int, str) -> Tuple[Union[float, np.array], Union[float, np.array]]
    """ Run t-test. Enable setting different alternative hypothesis.
    Probabilities are exact due to symmetry of the test.

    :return: (statistics, p_values)

    See also
    --------
    scipy.stats.ttest_ind

    """
    # alt = kwargs.get("alternative", ALT_TWO)
    assert alternative in ALTERNATIVES
    scores, pvalues = scipy.stats.ttest_ind(a, b, axis=axis)

    if alternative == ALT_TWO:
        return scores, pvalues

    less = scores < 0
    pvalues = pvalues / 2.0
    pvalues[np.logical_not(less)] = 1.0 - pvalues[np.logical_not(less)]

    if alternative == ALT_LESS:
        return scores, pvalues
    else:
        return scores, 1.0 - pvalues


def score_mann_whitney(a, b, **kwargs):
    axis = kwargs.get('axis', 0)
    a, b = np.asarray(a, dtype=float), np.asarray(b, dtype=float)

    if not 0 <= axis < 2:
        raise ValueError("Axis")

    if a.ndim != b.ndim:
        raise ValueError

    if axis >= a.ndim:
        raise ValueError

    if axis == 0:
        a, b = a.T, b.T

    alt = kwargs.get("alternative", ALT_TWO)
    assert alt in ALTERNATIVES
    statistics = np.zeros((a.shape[0]),)
    p_values = np.zeros((a.shape[0]),)
    for i, (a_, b_) in enumerate(zip(a, b)):
        try:
            s, p = scipy.stats.mannwhitneyu(a_, b_, alternative=alt)
        except ValueError:
            s, p = (0, 1)
        statistics[i] = s
        p_values[i] = p
    return np.array(statistics), np.array(p_values)


def score_hypergeometric_test(a, b, threshold=1, **kwargs):
    """
    Run a hypergeometric test. The probability in a two-sided test is approximated
    with the symmetric distribution with more extreme of the tails.
    """
    # type: (np.ndarray, np.ndarray, float) -> np.ndarray

    # Binary expression matrices
    A = (a >= threshold).astype(int)
    B = (b >= threshold).astype(int)
    alt = kwargs.get("alternative", ALT_TWO)
    assert alt in ALTERNATIVES

    # Test Parameters
    M = len(A) + len(B)
    N = len(A)
    n_expr = A.sum(axis=0) + B.sum(axis=0)  # Number of cells expressing genes (overall)
    n_expr_clust = A.sum(axis=0)            # Number of cells expressing genes (in cluster)

    # Test results --- both tails
    # Note: cumulatives do sum to >1 due to overlap at 1 point
    under = np.fromiter(map(lambda t: hypergeom.cdf(k=t[1], n=t[0], M=M, N=N),
                            zip(n_expr, n_expr_clust)), dtype=float)
    over = np.fromiter(map(lambda t: hypergeom.sf(k=t[1]-1, n=t[0], M=M, N=N),
                           zip(n_expr, n_expr_clust)), dtype=float)
    signs = np.sign(under - over)
    if alt == ALT_TWO:
        pvalues = np.minimum(1.0, 2.0 * np.minimum(under, over))
    elif alt == ALT_LESS:
        pvalues = under
    else:
        pvalues = over
    scores = -np.log(pvalues) * signs
    return scores, pvalues


def score_fold_change(a, b, axis=0, log=False):
    # type: (np.array, np.array, int, bool) -> np.array
    """ Calculate the fold change between `a` and `b` samples.

    :param a: Array containing the samples
    :param b: Array containing the samples
    :param axis: Axis over which to compute the scores
    :param log: Return the log2(scores).

    :return: The fold change scores
    """

    scores = np.nanmean(a, axis=axis) / np.nanmean(b, axis=axis)

    # TODO: Properly handle this warrning in widgets
    # "Negative fold change scores were ignored. You should use another scoring method."
    if np.any(scores < 0):
        scores[scores < 0] = np.nan

    return np.log2(scores) if log else scores


def _lngamma(z):
    x = 0
    x += 0.1659470187408462e-06 / (z + 7)
    x += 0.9934937113930748e-05 / (z + 6)
    x -= 0.1385710331296526 / (z + 5)
    x += 12.50734324009056 / (z + 4)
    x -= 176.6150291498386 / (z + 3)
    x += 771.3234287757674 / (z + 2)
    x -= 1259.139216722289 / (z + 1)
    x += 676.5203681218835 / z
    x += 0.9999999999995183
    
    return math.log(x) - 5.58106146679532777 - z + (z - 0.5) * math.log(z + 6.5)
        

class LogBin(object):
    _max = 2
    _lookup = [0.0, 0.0]
    _max_factorial = 1
    _lock = threading.Lock()

    def __init__(self, max=1000):
        self._extend(max)

    @staticmethod
    def _extend(max):
        with LogBin._lock:
            if max <= LogBin._max:
                return
            for i in range(LogBin._max, max):
                if i > 1000:  # an arbitrary cuttof
                    LogBin._lookup.append(LogBin._logfactorial(i))
                else:
                    LogBin._max_factorial *= i
                    LogBin._lookup.append(math.log(LogBin._max_factorial))
            LogBin._max = max

    def _logbin(self, n, k):
        if n >= self._max:
            self._extend(n + 100)
        if n > k >= 0:
            return self._lookup[n] - self._lookup[n - k] - self._lookup[k]
        else:
            return 0.0

    @staticmethod
    def _logfactorial(n):
        if (n <= 1):
            return 0.0
        else:
            return _lngamma(n + 1)


class Binomial(LogBin):
    """ `Binomial distribution <http://en.wikipedia.org/wiki/Binomial_distribution>`_ is a discrete
    probability distribution of the number of successes in a sequence
    of n independent yes/no experiments, each of which yields success
    with probability p. """

    def __call__(self, k, N, m, n):
        """ If m out of N experiments are positive return the probability
        that k out of n experiments are positive using the binomial
        distribution: if p = m/N then return bin(n,k)*(p**k + (1-p)**(n-k))
        where bin is the binomial coefficient. """

        p = 1.0 * m / N
        if p == 0.0:
            if k == 0:
                return 1.0
            else:
                return 0.0
        elif p == 1.0:
            if n == k:
                return 1.0
            else:
                return 0.0
        try:
            return min(math.exp(self._logbin(n, k) + k * math.log(p) + (n - k) * math.log(1.0 - p)), 1.0)
        except (OverflowError, ValueError) as er:
            print(k, N, m, n)
            raise
            # return math.exp(self._logbin(n, k) + math.log((p**k) * (1.0 - p)**(n - k)))

    def p_value(self, k, N, m, n):
        """ The probability that k or more tests are positive. """
        if n - k + 1 <= k:
            # starting from k gives the shorter list of values
            return sum(self.__call__(i, N, m, n) for i in range(k, n+1))
        else:
            value = 1.0 - sum(self.__call__(i, N, m, n) for i in range(k))
            # if the value is small it is probably inexact due to the limited
            # precision of floats, as for example  (1-(1-1e-20)) -> 0
            # if so, compute the result without substraction
            if value < 1e-3:  # arbitary threshold
                # print "INEXACT", value, sum(self.__call__(i, N, m, n) for i in range(k, n+1))
                return sum(self.__call__(i, N, m, n) for i in range(k, n+1))
            else:
                return value


class Hypergeometric(LogBin):
    """ `Hypergeometric distribution <http://en.wikipedia.org/wiki/Hypergeometric_distribution>`_ is
    a discrete probability distribution that describes the number of successes in a sequence of n draws
    from a finite population without replacement.
    """

    def __call__(self, k, N, m, n):
        """ If m out of N experiments are positive return the probability
        that k out of n experiments are positive using the hypergeometric
        distribution (i.e. return bin(m, k)*bin(N-m, n-k)/bin(N,n)
        where bin is the binomial coefficient).
        """
        if k < max(0, n + m - N) or k > min(n, m):
            return 0.0
        try:
            return min(math.exp(self._logbin(m, k) + self._logbin(N - m, n - k) - self._logbin(N, n)), 1.0)
        except (OverflowError, ValueError) as er:
            print(k, N, m, n)
            raise

    def p_value(self, k, N, m, n):
        """ 
        The probability that k or more tests are positive.
        """

        if min(n, m) - k + 1 <= k:
            # starting from k gives the shorter list of values
            return sum(self.__call__(i, N, m, n) for i in range(k, min(n,m)+1))
        else:
            value = 1.0 - sum(self.__call__(i, N, m, n) for i in (range(k)))
            # if the value is small it is probably inexact due to the limited
            # precision of floats, as for example  (1-(1-1e-20)) -> 0
            # if so, compute the result without substraction
            if value < 1e-3:  # arbitary threshold
                # print "INEXACT", value, sum(self.__call__(i, N, m, n) for i in range(k, min(n,m)+1))
                return sum(self.__call__(i, N, m, n) for i in range(k, min(n, m)+1))
            else:
                return value


# to speed-up FDR, calculate ahead sum([1/i for i in range(1, m+1)]), for m in [1,100000].
# For higher values of m use an approximation, with error less or equal to
# 4.99999157277e-006. (sum([1/i for i in range(1, m+1)])  ~ log(m) + 0.5772..., 0.5572 is an Euler-Mascheroni constant)
c = [1.0]
for m in range(2, 100000):
    c.append(c[-1] + 1.0/m)


def is_sorted(l):
    return all(l[i] <= l[i+1] for i in range(len(l)-1))


def FDR(p_values, dependent=False, m=None, ordered=False):
    """ `False Discovery Rate <http://en.wikipedia.org/wiki/False_discovery_rate>`_ correction on a list of p-values.

    :param p_values: a list of p-values.
    :param dependent: use correction for dependent hypotheses (default False).
    :param m: number of hypotheses tested (default ``len(p_values)``).
    :param ordered: prevent sorting of p-values if they are already sorted (default False).
    """

    if not ordered:
        ordered = is_sorted(p_values)

    if not ordered:
        joined = [(v, i) for i, v in enumerate(p_values)]
        joined.sort()
        p_values = [p[0] for p in joined]
        indices = [p[1] for p in joined]

    if not m:
        m = len(p_values)
    if m <= 0 or not p_values:
        return []

    if dependent: # correct q for dependent tests
        k = c[m-1] if m <= len(c) else math.log(m) + 0.57721566490153286060651209008240243104215933593992
        m = m * k

    tmp_fdrs = [p*m/(i+1.0) for (i, p) in enumerate(p_values)]
    fdrs = []
    cmin = tmp_fdrs[-1]
    for f in reversed(tmp_fdrs):
        cmin = min(f, cmin)
        fdrs.append( cmin)
    fdrs.reverse()

    if not ordered:
        new = [None] * len(fdrs)
        for v, i in zip(fdrs, indices):
            new[i] = v
        fdrs = new

    return fdrs


def Bonferroni(p_values, m=None):
    """ `Bonferroni correction <http://en.wikipedia.org/wiki/Bonferroni_correction>`_ correction on a list of p-values.

    :param p_values: a list of p-values.
    :param m: number of hypotheses tested (default ``len(p_values)``).
    """
    if not m:
        m = len(p_values)
    if m == 0:
        return []
    m = float(m)
    return [p/m for p in p_values]



"""
Caching framework for cached kegg api calls.

"""
import os
import sqlite3
try:
    import cPickle as pickle
except ImportError:
    import pickle

from contextlib import closing

from datetime import datetime, date, timedelta
from . import conf

import six

try:
    from UserDict import DictMixin
except ImportError:
    from collections import MutableMapping as DictMixin


class Store(object):
    def __init__(self):
        self.timestamp = 0

    def open(self):
        raise NotImplementedError

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


class Sqlite3Store(Store, DictMixin):

    def __init__(self, filename):
        self.filename = filename
        self.con = sqlite3.connect(filename)
        self.con.execute("""
            CREATE TABLE IF NOT EXISTS cache
                (key TEXT UNIQUE,
                 value TEXT
                )
        """)
        self.con.execute("""
            CREATE INDEX IF NOT EXISTS cache_index
            ON cache (key)
        """)
        self.con.commit()

    def __getitem__(self, key):
        cur = self.con.execute("""
            SELECT value
            FROM cache
            WHERE key=?
        """, (key,))
        r = cur.fetchall()
        if not r:
            raise KeyError(key)
        else:
            pickle_str = r[0][0]
            if not six.PY3:
                pickle_str = str(pickle_str)
            try:
                return pickle.loads(pickle_str)
            except Exception:
                raise KeyError(key)

    def __setitem__(self, key, value):
        value = pickle.dumps(value)
        self.con.execute("""
            INSERT OR REPLACE INTO cache
            VALUES (?, ?)
        """, (key, value))
        self.con.commit()

    def __delitem__(self, key):
        self.con.execute("""
            DELETE FROM cache
            WHERE key=?
        """, (key,))
        self.con.commit()

    def keys(self):
        cur = self.con.execute("""
            SELECT key
            FROM cache
        """)
        return [str(r[0]) for r in cur.fetchall()]

    def close(self):
        pass

    def __len__(self):
        return None

    def __iter__(self):
        return None



class DictStore(Store, DictMixin):
    def __init__(self):
        Store.__init__(self)

    def close(self):
        pass

    def __len__(self):
        return None

    def __iter__(self):
        return None



class cache_entry(object):
    def __init__(self, value, mtime=None, expires=None):
        self.value = value
        self.mtime = mtime
        self.expires = expires

_SESSION_START = datetime.now()


class cached_wrapper(object):
    """
    TODO: needs documentation
    """
    def __init__(self, function, instance, class_, cache_store,
                 last_modified=None):
        self.function = function
        self.instance = instance
        self.class_ = class_
        self.cache_store = cache_store
        self.last_modified = last_modified

    def has_key(self, key):
        with closing(self.cache_store()) as store:
            return key in store

    def key_from_args(self, args, kwargs=None):
        key = self.function.__name__ + repr(args)
        if kwargs is not None:
            raise NotImplementedError("kwargs are not supported yet")
        return key

    def invalidate_key(self, key):
        with closing(self.cache_store()) as store:
            del store[key]

    def last_modified_from_args(self, args, kwargs=None):
        if self.instance is not None:
            return self.instance.last_modified(args)

    def invalidate_args(self, args):
        return self.invalidate_key(self.key_from_args(args))

    def invalidate_all(self):
        prefix = self.key_from_args(()).rstrip(",)")
        with self.cache_store() as store:
            for key in store:
                if key.startswith(prefix):
                    del store[key]

    def memoize(self, args, kwargs, value, timestamp=None):
        key = self.key_from_args(args, kwargs)
        if timestamp is None:
            timestamp = datetime.now()

        with closing(self.cache_store()) as store:
            store[key] = cache_entry(value, mtime=timestamp)

    def __call__(self, *args):
        key = self.key_from_args(args)
        with closing(self.cache_store()) as store:
            if self.key_has_valid_cache(key, store):
                rval = store[key].value
            else:
                rval = self.function(self.instance, *args)
                store[key] = cache_entry(rval, datetime.now(), None)

        return rval

    def key_has_valid_cache(self, key, store):
        if key not in store:
            return False
        else:
            entry = store[key]
            return self.is_entry_valid(entry, None)

    def is_entry_valid(self, entry, args):
        # For now always return True, the caching architecture needs to be
        # completely reworked
        return True

        # Need to check datetime first (it subclasses date)
        if isinstance(entry.mtime, datetime):
            mtime = entry.mtime
        elif isinstance(entry.mtime, date):
            mtime = datetime(entry.mtime.year, entry.mtime.month,
                             entry.mtime.day, 1, 1, 1)
        else:
            return False

        last_modified = self.last_modified_from_args(args)

        if isinstance(last_modified, date):
            last_modified = datetime(last_modified.year, last_modified.month,
                                     last_modified.day, 1, 1, 1)
        elif isinstance(last_modified, str):
            # Could have different format
            mtime = mtime.strftime("%Y %m %d %H %M %S")

        elif last_modified is None:
            if conf.params["cache.invalidate"] == "always":
                return False
            elif conf.params["cache.invalidate"] == "session":
                last_modified = _SESSION_START
            elif conf.params["cache.invalidate"] == "daily":
                last_modified = datetime.now().replace(hour=0, minute=0,
                                                       second=0, microsecond=0)
            elif conf.params["cache.invalidate"] == "weekly":
                last_modified = datetime.now() - timedelta(7)
            else:  # ???
                pass
        return last_modified <= mtime


class cached_method(object):
    def __init__(self, function):
        self.function = function

    def __get__(self, instance, owner):
        if instance is not None:
            return cached_wrapper(self.function, instance, owner,
                                  self.get_cache_store(instance, owner))
        return self

    def get_cache_store(self, instance, owner):
        if hasattr(instance, "cache_store"):
            return instance.cache_store
        elif not hasattr("_cached_method_cache"):
            instance._cached_method_cache = DictStore()
        return instance._cached_method_cache


class bget_cached_method(cached_method):
    def __get__(self, instance, owner):
        if instance is not None:
            return cached_wrapper(self.function, instance, owner,
                                  self.get_cache_store(instance, owner),
                                  self.get_last_modified(instance, owner))
        return self

    def get_last_modified(self, instance, owner):
        if hasattr(instance, "last_modified"):
            return instance.last_modified


def touch_dir(path):
    path = os.path.expanduser(path)
    if not os.path.exists(path):
        os.makedirs(path)


def clear_cache():
    """Clear all locally cached KEGG data.
    """
    import glob
    path = conf.params["cache.path"]
    if os.path.realpath(path) != os.path.realpath(conf.kegg_dir):
        raise Exception("Non default cache path. Please remove the contents "
                        "of %r manually." % path)

    for cache_filename in glob.glob(os.path.join(path, "*.sqlite3")):
        os.remove(cache_filename)

    for ko_filename in glob.glob(os.path.join(path, "*.keg")):
        os.remove(ko_filename)

    for kgml_filename in glob.glob(os.path.join(path, "*.xml")):
        os.remove(kgml_filename)

    for png_filename in glob.glob(os.path.join(path, "*.png")):
        os.remove(png_filename)



import numpy as np
from Orange.data import Domain, ContinuousVariable, Table

from orangecontrib.bioinformatics.ncbi.gene import GeneInfo
from orangecontrib.bioinformatics.utils import statistics
from scipy.stats.mstats import rankdata
from scipy.stats import hypergeom, binom

from orangecontrib.bioinformatics.widgets.utils.data import TAX_ID

SCORING_EXP_RATIO = "scoring_exp_ratio"
SCORING_MARKERS_SUM = "scoring_sum_of_expressed_markers"
SCORING_LOG_FDR = "scoring_log_fdr"
SCORING_LOG_PVALUE = "scoring_log_p_value"

PFUN_BINOMIAL = "binomial_p_function"
PFUN_HYPERGEOMETRIC = "HYPERGEOMETRIC_p_function"


class ScoringNotImplemented(Exception):
    pass


class AnnotateSamples:
    """
    AnnotateSamples is class used for the annotation of data items with the
    labels Mann-Whitney U test for selecting important values and
    the Hyper-geometric for assigning the labels.

    Example for full annotation:

    >>> from Orange.data import Table
    >>> from orangecontrib.bioinformatics.utils import serverfiles
    >>> from orangecontrib.bioinformatics.annotation.annotate_samples import \
    ...     AnnotateSamples
    >>> from orangecontrib.bioinformatics.widgets.utils.data import TAX_ID
    >>>
    >>> data = Table("https://datasets.orange.biolab.si/sc/aml-1k.tab.gz")
    >>> data.attributes[TAX_ID] = "9606"  # table needs to have an organism ID
    >>> markers_path = serverfiles.localpath_download(
    ...     'marker_genes','panglao_gene_markers.tab')
    >>> marker = Table(markers_path)
    >>>
    >>> # filter only human markers
    >>> from Orange.data.filter import FilterString, Values
    >>> f = FilterString("Organism", FilterString.Equal, "Human")
    >>> markers = Values([f])(marker)
    >>>
    >>> annotator = AnnotateSamples()
    >>> annotations = annotator.annotate_samples(
    ...     data, markers, p_threshold=0.05)

    Example for full manual annotation. Here annotation is split in three
    phases. We assume that data are already loaded.

    >>> annotator = AnnotateSamples()
    >>> z = annotator.mann_whitney_test(data)
    >>> scores, p_val = AnnotateSamples.assign_annotations(z, markers, data)
    >>> scores = annotator.filter_annotations(scores, p_val, p_threshold=0.05)

    Attributes
    ----------

    """
    @staticmethod
    def log_cpm(data):
        """
        Function normalizes data with CPM methods and normalize them.

        Parameters
        ----------
        data : Orange.data.Table
            Tabular data with gene expressions

        Returns
        -------
        Orange.data.Table
            Normalized gene expression data
        """
        norm_data = np.log(1 + AnnotateSamples._cpm(data.X))
        return Table(data.domain, norm_data)

    @staticmethod
    def _cpm(data):
        """
        This function normalizes data with CPM methods.

        Parameters
        ----------
        data : array_like
            Numpy array with data. Columns are genes, rows are cells.
        """
        return data / np.sum(data, axis=1)[:, None] * 1e6

    @staticmethod
    def _ranks(data):
        """
        This function computes ranks for data in the table along axis=0.
        """
        x_len = data.shape[0]
        x_mask = data.sum(axis=0) > 0

        # create a matrix of ranges - init with average rank
        # for columns without nonzero expressions
        data_ge_ranked = np.ones(data.shape) * (1 + data.shape[0]) / 2

        # compute ranks only for nonzero columns
        for i in np.where(x_mask)[0]:
            mask = data[:, i] > 0
            col = np.ones(x_len) * (1 + (x_len - mask.sum())) / 2
            col[mask] = rankdata(data[mask, i]) + (x_len - mask.sum())
            data_ge_ranked[:, i] = col
        return data_ge_ranked

    @staticmethod
    def mann_whitney_test(data):
        """
        Compute z values with test Mann-Whitney U test.

        Parameters
        ----------
        data : Orange.data.Table
            Tabular data with gene expressions

        Returns
        -------
        Orange.data.Table
            Z-value for each item.
        """
        if len(data.X) <= 1:
            return [], []

        # rank data
        data_ge_ranked = AnnotateSamples._ranks(data.X)

        # compute U, mu, sigma
        n = data_ge_ranked.shape[0]
        n2 = n - 1
        u = data_ge_ranked - 1
        mu = n2 / 2
        sigma = np.zeros(data_ge_ranked.shape[1])

        for i in range(data_ge_ranked.shape[1]):
            _, counts = np.unique(data_ge_ranked[:, i], return_counts=True)
            sigma[i] = np.sqrt(
                1 * n2 / 12 * ((n + 1) - np.sum((counts ** 3 - counts)) /
                               (n * (n - 1))))

        # compute z
        z = (u - mu) / (sigma + 1e-16)

        # pack z values to data table
        # take only attributes in new domain
        domain = Domain([x for x in data.domain.attributes])
        z_table = Table(domain, z)

        return z_table

    @staticmethod
    def _reorder_matrix(matrix, genes_order):
        """
        Function reorder the columns of the array to fit to the genes_order

        Parameters
        ----------
        matrix : Orange.data.Table
            Tabular data tha needs to be reordered
        genes_order : list
            Desired genes order

        Returns
        ------
        np.ndarray
            Reordered array.
        """
        current_order = np.array(
            [x.attributes.get("Entrez ID")
             for x in matrix.domain.attributes])
        values = matrix.X

        # filter out genes without entrez ID
        has_entrez_id = current_order != None
        # just in case if Entrez ID not strings
        current_order = np.array([str(x) for x in current_order[has_entrez_id]])
        values = values[:, has_entrez_id]

        genes_order = np.array(genes_order)

        xsorted = np.argsort(genes_order)
        ypos = np.searchsorted(genes_order[xsorted], current_order)
        indices = xsorted[ypos]  # index which tell where should be the column

        reordered_values = np.zeros((values.shape[0], len(genes_order)))
        for i_curr, i_dest in enumerate(indices):
            reordered_values[:, i_dest] = values[:, i_curr]

        return reordered_values

    @staticmethod
    def _select_attributes(z, genes_order, z_threshold=1):
        """
        Function selects "over"-expressed attributes for items based on z
        values. It also reorder the matrix columns.

        Parameters
        ----------
        z : Orange.data.Table
            Tabular data z values for each item in the table
        genes_order : list
            Desired genes order
        z_threshold : float
            The threshold for selecting the attribute. For each item the
            attributes with z-value above this value are selected.

        Returns
        -------
        np.ndarray
            Reordered and thresholded z-values/
        """
        reordered_z = AnnotateSamples._reorder_matrix(z, genes_order)

        return reordered_z > z_threshold

    @staticmethod
    def _group_marker_attributes(markers, genes_order):
        """
        Function transforms annotations table to dictionary with format
        {annotation1: [attributes], annotation2: [attributes], ...}
        """
        types = sorted(list(set(markers[:, "Cell Type"].metas.flatten())))
        genes_celltypes = np.zeros((len(genes_order), len(types)))

        for m in markers:
            g = str(m["Entrez ID"].value)
            m = m["Cell Type"].value
            if g is not None and not g == "?":
                genes_celltypes[genes_order.index(g), types.index(m)] = 1

        return genes_celltypes, types

    @staticmethod
    def _score(scoring_type, p_values, fdrs, data, M, x, m, genes_order):
        if scoring_type == SCORING_MARKERS_SUM:
            return AnnotateSamples._reorder_matrix(data, genes_order).dot(M)
        if scoring_type == SCORING_EXP_RATIO:
            return x / m
        if scoring_type == SCORING_LOG_FDR:
            return -np.log(fdrs)
        if scoring_type == SCORING_LOG_PVALUE:
            return -np.log(p_values)
        else:
            raise ScoringNotImplemented()

    @staticmethod
    def assign_annotations(z_values, available_annotations, data, z_threshold=1,
                           p_value_fun=PFUN_BINOMIAL,
                           scoring=SCORING_EXP_RATIO):
        """
        The function gets a set of attributes (e.g. genes) for each cell and
        attributes for each annotation. It returns the annotations significant
        for each cell.

        Parameters
        ----------
        z_values : Orange.data.Table
            Table which show z values for each item
        available_annotations : Orange.data.Table
            Available annotations (e.g. cell types)
        z_threshold : float
            The threshold for selecting the attribute. For each item the
            attributes with z-value above this value are selected.
        p_value_fun : str, optional (defaults: TEST_BINOMIAL)
            A function that calculates p-value. It can be either
            PFUN_BINOMIAL that uses statistics.Binomial().p_value or
            PFUN_HYPERGEOMETRIC that uses hypergeom.sf.
        data : Orange.data.Table
            Tabular data with gene expressions - we need that to compute scores.
        scoring : str, optional (default=SCORING_EXP_RATIO)
            Type of scoring

        Returns
        -------
        Orange.data.Table
            Annotation probabilities
        Orange.data.Table
            Annotation fdrs
        """
        assert TAX_ID in data.attributes, "The input table needs to have a " \
                                          "tax_id attribute"
        tax_id = data.attributes[TAX_ID]

        # select function for p-value
        if p_value_fun == PFUN_HYPERGEOMETRIC:
            p_fun = lambda x, N, m, k: hypergeom.sf(x, N, m, k)
        else:
            p_fun = lambda x, N, m, k: binom.sf(x, k, m / N)

        N = len(GeneInfo(tax_id))  # number of genes for organism

        # make an attributes order
        genes_data = [str(x.attributes["Entrez ID"])
                      for x in z_values.domain.attributes
                      if "Entrez ID" in x.attributes]
        genes_celltypes = [
            str(x) for x in available_annotations[:, "Entrez ID"].metas.flatten()
            if x is not None and not x == "?"]
        genes_order = list(set(genes_data) | set(genes_celltypes))

        # get marker genes matrix M
        M, annotations = AnnotateSamples._group_marker_attributes(
            available_annotations, genes_order)

        Z = AnnotateSamples._select_attributes(
            z_values, genes_order, z_threshold)

        x = Z.dot(M)
        k = np.repeat(Z.sum(axis=1).reshape(-1, 1), x.shape[1], axis=1)
        m = np.repeat(M.sum(axis=0).reshape(1, -1), x.shape[0], axis=0)

        p_values = p_fun(x - 1, N, m, k)

        fdrs = np.zeros(p_values.shape)
        for i, row in enumerate(p_values):
            fdrs[i] = np.array(statistics.FDR(row.tolist()))

        scores = AnnotateSamples._score(
            scoring, p_values, fdrs, data, M, x, m, genes_order)

        domain = Domain(
            [ContinuousVariable(ct) for ct in annotations])
        scores_table = Table(domain, scores)
        fdrs_table = Table(domain, fdrs)

        return scores_table, fdrs_table

    @staticmethod
    def filter_annotations(scores, p_values, return_nonzero_annotations=True,
                           p_threshold=0.05):
        """
        This function filters the probabilities on places that do not reach the
        threshold for p-value and filter zero columns
        return_nonzero_annotations is True.

        Parameters
        ----------
        scores : Orange.data.Table
            Scores for each annotations for each cell
        p_values : Orange.data.Table
            p-value scores for annotations for each cell
        return_nonzero_annotations : bool
            Flag that enables filtering the non-zero columns.
        p_threshold : float
            A threshold for accepting the annotations. Annotations that has FDR
            value bellow this threshold are used.


        Returns
        -------
        Orange.data.Table
            Filtered scores for each annotations for each cell
        """
        scores_x = np.copy(scores.X)  # do not want to edit values inplace
        scores_x[p_values.X > p_threshold] = np.nan
        scores = Table(scores.domain, scores_x)

        if return_nonzero_annotations:
            col_not_empty = ~np.isnan(scores).all(axis=0)
            new_domain = Domain(
                np.array(scores.domain.attributes)[col_not_empty])
            scores = Table(new_domain, scores)
        return scores

    @staticmethod
    def annotate_samples(data, available_annotations,
                         return_nonzero_annotations=True, p_threshold=0.05,
                         p_value_fun=PFUN_BINOMIAL, z_threshold=1,
                         scoring=SCORING_EXP_RATIO, normalize=False):
        """
        Function marks the data with annotations that are provided. This
        function implements the complete functionality. First select genes,
        then annotate them and filter them.

        Parameters
        ----------
        data : Orange.data.Table
            Tabular data
        available_annotations : Orange.data.Table
            Available annotations (e.g. cell types)
        return_nonzero_annotations : bool, optional (default=True)
            If true return scores for only annotations present in at least one
            sample.
        p_threshold : float
            A threshold for accepting the annotations. Annotations that has FDR
            value bellow this threshold are used.
        p_value_fun : str, optional (defaults: TEST_BINOMIAL)
            A function that calculates p-value. It can be either
            PFUN_BINOMIAL that uses statistics.Binomial().p_value or
            PFUN_HYPERGEOMETRIC that uses hypergeom.sf.
        z_threshold : float
            The threshold for selecting the attribute. For each item the
            attributes with z-value above this value are selected.
        scoring : str, optional (default = SCORING_EXP_RATIO)
            Type of scoring
        normalize : bool, optional (default = False)
            This variable tells whether to normalize data or not.

        Returns
        -------
        Orange.data.Table
            Cell type most important for each cell.
        """
        assert len(data) > 1, "At least two data items are required for " \
                              "method to work."

        if normalize:
            data = AnnotateSamples.log_cpm(data)

        z = AnnotateSamples.mann_whitney_test(
            data)

        annotation_probs, annotation_fdrs = AnnotateSamples.assign_annotations(
            z, available_annotations, data, z_threshold=z_threshold,
            p_value_fun=p_value_fun, scoring=scoring)

        annotation_probs = AnnotateSamples.filter_annotations(
            annotation_probs, annotation_fdrs, return_nonzero_annotations,
            p_threshold
        )

        return annotation_probs



""" GeneInfo utils """
import os
import sqlite3

from contextlib import closing

from orangecontrib.bioinformatics.ncbi.gene import DOMAIN, FILENAME
from orangecontrib.bioinformatics.utils import serverfiles


class GeneInfoFileNotFound(Exception):
    pass


def parse_synonyms(values):
    return [val for val in values.split('|') if val]


def parse_sources(values):
    """ Parse source string from NCBI gene info.

    ftp://ftp.ncbi.nlm.nih.gov/gene/DATA/README:

    bar-delimited set of identifiers in other databases
    for this gene.  The unit of the set is database:value.
    Note that HGNC and MGI include 'HGNC' and 'MGI', respectively,
    in the value part of their identifier.

    Consequently, dbXrefs for these databases will appear like: HGNC:HGNC:1100
    This would be interpreted as database='HGNC', value='HGNC:1100'


    Args:
        values (str): string of gene sources

    Returns:
        :obj:`dict`: Keys are source names, values are source ids

    """
    external_ids = values.split('|')
    out_dict = {}

    for ref in external_ids:
        if ref != '-':

            source_dest, source_id = ref.split(':', 1)
            out_dict[source_dest] = source_id

    return out_dict


class GeneInfoDB:

    def __init__(self):
        db_path = serverfiles.localpath_download(DOMAIN, FILENAME)

        if os.path.isfile(db_path):
            self._db_con = sqlite3.connect(db_path)
        else:
            raise GeneInfoFileNotFound(db_path)

    def __len__(self):
        with closing(self._db_con.cursor()) as cursor:
            return cursor.execute('SELECT COUNT(gene_id) FROM gene_info').fetchone()[0]

    def select_gene_info(self, gene_id):
        with closing(self._db_con.cursor()) as cursor:
            return cursor.execute('SELECT tax_id, gene_id, symbol, synonyms, db_refs, description, locus_tag,'
                                  'chromosome, map_location, type_of_gene, symbol_from_nomenclature_authority,'
                                  'full_name_from_nomenclature_authority, nomenclature_status, other_designations,'
                                  'modification_date FROM gene_info '
                                  'WHERE gene_id = ?', (gene_id,)).fetchone()

    def select_genes_by_organism(self, organism):
        with closing(self._db_con.cursor()) as cursor:
            return cursor.execute('SELECT tax_id, gene_id, symbol, synonyms, db_refs, description, locus_tag,'
                                  'chromosome, map_location, type_of_gene, symbol_from_nomenclature_authority,'
                                  'full_name_from_nomenclature_authority, nomenclature_status, other_designations,'
                                  'modification_date FROM gene_info '
                                  'WHERE species = ? or tax_id = ?', (organism, organism)).fetchall()

    def select_gene_matcher_data(self, organism):
        with closing(self._db_con.cursor()) as cursor:
            return cursor.execute('SELECT tax_id, gene_id, symbol, synonyms, db_refs, locus_tag, '
                                  'symbol_from_nomenclature_authority FROM gene_info '
                                  'WHERE species = ? or tax_id = ?', (organism, organism)).fetchall()


if __name__ == "__main__":

    def main():
        test = GeneInfoDB()
        genes = test.select_gene_matcher_data("562")
        print(type(genes), len(genes), genes[0])
        sel_genes = test.select_genes_by_organism('562')
        print(type(sel_genes), len(sel_genes), sel_genes[0])

    main()



import numpy as np


from Orange.widgets.widget import OWWidget, Msg
from Orange.widgets import gui, settings
from Orange.widgets.settings import SettingProvider
from Orange.widgets.visualize.owscatterplot import OWScatterPlotBase,  OWDataProjectionWidget

from orangecontrib.bioinformatics.widgets.utils.gui import label_selection
from orangecontrib.bioinformatics.utils.statistics import score_fold_change, score_t_test
from orangecontrib.bioinformatics.widgets.utils.data import (
    GENE_AS_ATTRIBUTE_NAME, GENE_ID_COLUMN
)


class VolcanoGraph(OWScatterPlotBase):
    label_only_selected = settings.Setting(True)

    def set_axis_title(self, axis, title):
        self.plot_widget.setLabel(axis=axis, text=title)


class OWVolcanoPlot(OWDataProjectionWidget):
    name = "Volcano Plot"
    description = "Plots fold change vs. p-value."
    icon = "icons/OWVolcanoPlot.svg"
    priority = 100

    class Warning(OWDataProjectionWidget.Warning):
        insufficient_data = Msg('Insufficient data to compute statistics.'
                                'More than one measurement per class should be provided ')

        gene_enrichment = Msg('{}, {}.')
        no_selected_gene_sets = Msg('No gene set selected, select them from Gene Sets box.')

    class Error(OWDataProjectionWidget.Error):
        exclude_error = Msg('Target labels most exclude/include at least one value.')
        negative_values = Msg('Negative values in the input. The inputs cannot be in ratio scale.')
        data_not_annotated = Msg('The input date is not annotated as expexted. Please refer to documentation.')
        gene_column_id_missing = Msg('Can not identify genes column. Please refer to documentation.')

    GRAPH_CLASS = VolcanoGraph
    graph = SettingProvider(VolcanoGraph)
    embedding_variables_names = ('log2 (ratio)', '-log10 (P_value)')

    stored_selections = settings.ContextSetting([])
    current_group_index = settings.ContextSetting(0)

    def __init__(self):
        super().__init__()

    def _add_controls(self):
        box = gui.vBox(self.controlArea, "Target Labels")
        self.group_selection_widget = label_selection.LabelSelectionWidget()
        self.group_selection_widget.groupChanged.connect(self.on_target_values_changed)
        self.group_selection_widget.groupSelectionChanged.connect(self.on_target_values_changed)
        box.layout().addWidget(self.group_selection_widget)

        super()._add_controls()
        self.gui.add_widgets([self.gui.ShowGridLines], self._plot_box)

    def get_embedding(self):
        self.Error.exclude_error.clear()

        group, target_indices = self.group_selection_widget.selected_split()

        if self.data and group is not None and target_indices:
            X = self.data.X
            I1 = label_selection.group_selection_mask(self.data, group, target_indices)
            I2 = ~I1

            # print(group)
            if isinstance(group, label_selection.RowGroup):
                X = X.T

            N1, N2 = np.count_nonzero(I1), np.count_nonzero(I2)

            if not N1 or not N2:
                self.Error.exclude_error()
                return

            if N1 < 2 and N2 < 2:
                self.Warning.insufficient_data()

            X1, X2 = X[:, I1], X[:, I2]

            if np.any(X1 < 0.0) or np.any(X2 < 0):
                self.Error.negative_values()
                X1 = np.full_like(X1, np.nan)
                X2 = np.full_like(X2, np.nan)

            with np.errstate(divide="ignore", invalid="ignore"):
                fold = score_fold_change(X1, X2, axis=1, log=True)
                _, p_values = score_t_test(X1, X2, axis=1)
                log_p_values = np.log10(p_values)

            self.valid_data = np.isfinite(fold) & np.isfinite(p_values)
            return np.array([fold, -log_p_values]).T

    def setup_plot(self):
        super().setup_plot()
        for axis, var in (("bottom", 'log<sub>2</sub> (ratio)'), ("left", '-log<sub>10</sub> (P_value)')):
            self.graph.set_axis_title(axis, var)

    def on_target_values_changed(self, index):
        # Save the current selection to persistent settings
        self.current_group_index = index
        selected_indices = [ind.row() for ind in self.group_selection_widget.currentGroupSelection().indexes()]

        if self.current_group_index != -1 and selected_indices:
            self.stored_selections[self.current_group_index] = selected_indices

        self.setup_plot()

    def set_data(self, data):
        self.Warning.clear()
        self.Error.clear()
        super().set_data(data)
        self.group_selection_widget.set_data(self, self.data)

        if self.data:
            if not self.stored_selections:
                self.stored_selections = [[0] for _ in self.group_selection_widget.targets]
            self.group_selection_widget.set_selection()

    def check_data(self):
        self.clear_messages()
        use_attr_names = self.data.attributes.get(GENE_AS_ATTRIBUTE_NAME, None)
        gene_id_column = self.data.attributes.get(GENE_ID_COLUMN, None)

        if self.data is not None and (len(self.data) == 0 or
                                      len(self.data.domain) == 0):
            self.data = None

        if use_attr_names is None:
            # Note: input data is not annotated properly.
            self.Error.data_not_annotated()
            self.data = None

        if gene_id_column is None:
            # Note: Can not identify genes column.
            self.Error.gene_column_id_missing()
            self.data = None


if __name__ == "__main__":
    pass



""" GEO Dataset (GDS) """
from collections import OrderedDict
from orangecontrib.bioinformatics.geo import dataset_all_info, dataset_download


class GDSInfo(OrderedDict):
    __slots__ = ()

    def __init__(self):
        """ Retrieve information about `GEO DataSets <http://www.ncbi.nlm.nih.gov/sites/GDSbrowser>`_.

        The class accesses the Orange server file that either resides on the local computer or
        is automatically retrieved from Orange server. Calls to this class do not access any NCBI's servers.

        The constructor will download information on GEO DataSets that are curated and stored on Orange database
        servers.

        An instance behaves like a dictionary: the keys are GEO DataSets IDs, and the dictionary values
        for is a dictionary providing various information about the particular data set.

        Example
        --------
            >>> info = GDSInfo()
            >>> list(info.keys())[:5]
            ['GDS10', 'GDS100', 'GDS1001', 'GDS1002', 'GDS1003']
            >>> info['GDS10']['title']
            'Type 1 diabetes gene expression profiling'
            >>> info['GDS10']['platform_organism']
            'Mus musculus'

        """
        super().__init__()
        self.update({gds_info['name']: gds_info for _, gds_info in dataset_all_info()})


def get_samples(gds_info: dict):
    return {sample for subset in gds_info['subsets'] for sample in subset['sample_id']}


GDS = dataset_download

if __name__ == '__main__':
    info = GDSInfo()
    print(list(info.keys())[:5])
    print(info['GDS10']['title'])
    print(info['GDS10']['platform_organism'])

    dataset_download('GDS1001')



""" NCBI GeneInformation module """
import pickle

from collections import defaultdict
from functools import lru_cache
from typing import List, Union, Optional
from types import SimpleNamespace

from requests.exceptions import ConnectTimeout, RequestException, ConnectionError

from Orange.data import StringVariable, DiscreteVariable, Domain, Table


from orangecontrib.bioinformatics.ncbi.gene.config import *
from orangecontrib.bioinformatics.ncbi.gene.utils import GeneInfoDB, parse_sources, parse_synonyms
from orangecontrib.bioinformatics.utils import serverfiles, ensure_type
from orangecontrib.bioinformatics.widgets.utils.data import TAX_ID, GENE_ID_COLUMN, GENE_AS_ATTRIBUTE_NAME

_no_hits, _single_hit, _multiple_hits = 0, 1, 2
_source, _symbol, _synonym, _locus, _gene_id, _nom_symbol = \
    'External reference', 'Symbol', 'Synonym', 'Locus tag', 'NCBI ID', 'Nomenclature symbol'


class OrangeTableAnnotations(SimpleNamespace):
    tax_id: str = 'taxonomy_id'
    gene_id_column: str = 'gene_id_column'
    gene_id_attribute: str = 'gene_id_attribute'
    gene_as_attribute_name: str = 'gene_as_attribute_name'


class NoGeneNamesException(Exception):
    """ Can not extract gene names from the table. Table can't have both row and column gene variables """


class Gene:
    """ Basic representation of a gene. Class holds NCBI and GeneMatcher data created upon class initialization.

    GeneMatcher tags are used only for matching purposes.
    """

    # We deny __dict__ creation by defining __slots__ for memory optimization.
    # We expect a lot of instances of this class to be created.
    __slots__ = GENE_INFO_TAGS + GENE_MATCHER_TAGS  # Note: slots are ordered.

    def __init__(self, input_name=None):
        """
        :param input_name: Name of the gene (symbol, synonym, external reference)
        :type input_name: :class:`str`
        """
        self.input_name = input_name
        self.type_of_match = None
        self.ncbi_id = None
        self._possible_hits = []

    def __getattr__(self, attribute):
        if attribute not in self.__slots__:
            return ''

    @property
    def possible_hits(self):
        return self._possible_hits

    @possible_hits.setter
    def possible_hits(self, genes):
        for gene in genes:
            possible_match = Gene()
            possible_match.ncbi_id = gene
            self._possible_hits.append(possible_match)

    @lru_cache(10000)
    def load_ncbi_info(self):
        """ Populate :class:`Gene` with data from NCBI gene database
        """
        if not self.ncbi_id:
            return

        info = GeneInfoDB().select_gene_info(self.ncbi_id)
        if info:
            for attr, value in zip(self.__slots__, info):
                if attr == 'db_refs':
                    value = parse_sources(value)
                elif attr == 'synonyms':
                    value = parse_synonyms(value)

                setattr(self, attr, value)

    def to_list(self):
        _, header_tags = GENE_MATCHER_HEADER

        def parse_attribute(tag):
            gene_attr = getattr(self, '{}'.format(tag))

            if isinstance(gene_attr, dict):
                # note: db_refs are stored as dicts
                gene_attr = ', '.join('{}: {}'.format(key, val)
                                      for (key, val) in gene_attr.items()) if gene_attr else ' '
            elif isinstance(gene_attr, list):
                # note: synonyms are stored as lists
                gene_attr = ', '.join(gene_attr) if gene_attr else ' '

            return gene_attr

        return [parse_attribute(tag) for tag in header_tags]

    def to_html(self):
        """ Return a basic HTML representation of :class:`Gene` class.

        :rtype: :class:`str`
        """
        self.load_ncbi_info()
        db_refs = getattr(self, 'db_refs')
        external_links = []
        if db_refs:
            external_links = ['<dd>- <b>{}</b>: {}<dd>'.format(ref, ref_id) for ref, ref_id in db_refs.items()]

        html_string = """<span>
                      <h3><b>Gene info summary</b></h3>
                      <dl>
                      <dt><b>Gene ID:</b></dt>
                      <dd>- {}</dd>
                      
                      <dt><b>Symbol:</b></dt>
                      <dd>- {}</dd>
   
                      <dt><b>Synonyms:</b></dt>
                      <dd>- {}</dd> 
                      
                      <dt><b>External references:</b></dt>
                      {}
                      
                      <dt><b>Description:</b></dt>
                      <dd>{}</dd>
                      
                      <dt><b>Type of gene:</b></dt>
                      <dd>{}</dd>
                      </dl>
                      </span>""".format(self.ncbi_id,
                                        getattr(self, 'symbol'),
                                        ', '.join([synonym for synonym in getattr(self, 'synonyms')]),
                                        '\n'.join([link for link in external_links]),
                                        getattr(self, 'description'),
                                        getattr(self, 'type_of_gene'))

        return html_string


class GeneInfo(dict):
    def __init__(self, organism):
        """ Load genes for given organism.

        Each instance of :class:`Gene` is mapped to corresponding Entrez ID

        :param organism: Taxonomy if (NCBI taxonomy database)
        :type organism:  class:`str` or :class:`int`
        """
        super().__init__()
        self.__init_gene_info(organism)

    def __init_gene_info(self, organism):
        for gene in GeneInfoDB().select_genes_by_organism(organism):
            gene_obj = Gene()

            for attr, value in zip(gene_obj.__slots__, gene):
                if attr == 'db_refs':
                    value = parse_sources(value)
                elif attr == 'synonyms':
                    value = parse_synonyms(value)

                setattr(gene_obj, attr, value)

            self[gene_obj.gene_id] = gene_obj

    def get_gene_by_id(self, gene_id):
        """ Return the :class:`Gene` object for given gene_id

        :param gene_id: Entrez ID (NCBI gene database)
        :type gene_id: str or int

        :rtype: :class:`Gene`
        """

        try:
            return self[int(gene_id)]
        except KeyError as e:
            # TODO: handle this properly?
            # raise e
            pass


class GeneMatcher:

    def __init__(self,  tax_id, **kwargs):
        """ Gene name matching interface

        :param tax_id: Taxonomy if (NCBI taxonomy database)
        :type tax_id: str

        """
        self._organism = ensure_type(tax_id, str)  # type: str
        self._genes = []                           # type: (List[Union[str, Gene]])

        self._case_insensitive = kwargs.get("case_insensitive", False)
        self._matcher = self.load_matcher_file(DOMAIN, MATCHER_FILENAME.format(tax_id))

    @property
    def organism(self):
        return self._organism

    @organism.setter
    def organism(self, tax_id):
        self._organism = ensure_type(tax_id, str)
        self._matcher = self.load_matcher_file(DOMAIN, MATCHER_FILENAME.format(tax_id))

    @property
    def genes(self):
        return self._genes

    @genes.setter
    def genes(self, gene_data):  # type: (List[str]) -> None
        self._genes = [Gene(input_name=gene) for gene in gene_data]

    def get_known_genes(self):
        """ Return genes with known Entrez Id from NCBI gene database

        Unknown genes are not included.

        :rtype: :class:`list` of :class:`Gene` instances
        """
        return [gene for gene in self.genes if gene.ncbi_id]

    def map_input_to_ncbi(self):
        """ Return a map where key is a given input (gene name) and value is gene id (Entrez Id)

        Unknown genes are not included.

        :rtype: :class:`dict`
        """
        return {gene.input_name: gene.ncbi_id for gene in self.genes if gene.ncbi_id}

    def match_table_attributes(self, data_table):
        """ Helper function for gene name matching in data table.

        Match table attributes and if a unique match is found create a new column attribute for Entrez Id.
        Attribute name is defined here: `orangecontrib.bioinformatics.ncbi.gene.config.NCBI_ID`


        :param data_table: data table
        :type data_table: :class:`Orange.data.Table`

        """
        input_gene_names = [var.name for var in data_table.domain.attributes]

        if input_gene_names:
            self.genes = input_gene_names
            self.run_matcher()

            for gene in self.genes:
                if gene.ncbi_id:
                    data_table.domain[gene.input_name].attributes[NCBI_ID] = gene.ncbi_id

    def match_table_column(self, data_table: Table, column_name: str,
                           target_column: Optional[StringVariable] = None) -> Table:
        """ Helper function for gene name matching in data table.

        :param data_table: data table
        :param column_name: Name of the column where gene symbols are located
        :param target_column: Column
        :type data_table: :class:`Orange.data.Table`

        """

        if column_name in data_table.domain:
            self.genes = data_table.get_column_view(column_name)[0]
            self.run_matcher()

            if target_column is None:
                target_column = StringVariable(NCBI_ID)

            domain = Domain([], metas=[target_column])
            data = [[str(gene.ncbi_id) if gene.ncbi_id else '?'] for gene in self.genes]
            table = Table(domain, data)

            return Table.concatenate([data_table, table])

    @staticmethod
    def gene_match_status(gene):
        # type: (Gene) -> str

        if gene.ncbi_id:
            return 'Matched'
        elif gene.possible_hits:
            return 'Match Conflict'
        else:
            return 'Unmatched'

    def to_data_table(self, selected_genes=None):
        tax_id = set()
        data_x = []
        metas = [
            StringVariable('Input gene ID'),
            DiscreteVariable('Match result', values=['Matched', 'Match Conflict', 'Unmatched']),
            StringVariable(NCBI_ID),
            StringVariable('Symbol'),
            StringVariable('Synonyms'),
            StringVariable('Description'),
            StringVariable('Other IDs'),
            StringVariable('Type of gene'),
            StringVariable('Chromosome'),
            StringVariable('Map location'),
            StringVariable('Locus tag'),
            StringVariable('Symbol from nomenclature authority'),
            StringVariable('Full name from nomenclature authority'),
            StringVariable('Nomenclature status'),
            StringVariable('Other designations'),
            StringVariable('Taxonomy ID'),
        ]
        domain = Domain([], metas=metas)

        genes = self.genes
        if selected_genes is not None:
            genes = [gene for gene in self.genes if str(gene.ncbi_id) in selected_genes]

        for gene in genes:
            gene.load_ncbi_info()
            tax_id.add(gene.tax_id)
            match_status = self.gene_match_status(gene)

            db_refs = ', '.join('{}: {}'.format(key, val)
                                for (key, val) in gene.db_refs.items()) if gene.db_refs else ''
            synonyms = ', '.join(gene.synonyms) if gene.synonyms else ''

            line = [
                gene.input_name, match_status, gene.ncbi_id, gene.symbol, synonyms, gene.description, db_refs,
                gene.type_of_gene, gene.chromosome, gene.map_location, gene.locus_tag,
                gene.symbol_from_nomenclature_authority, gene.full_name_from_nomenclature_authority,
                gene.nomenclature_status, gene.other_designations, gene.tax_id
            ]

            data_x.append(line)

        tax_id = filter(None.__ne__, tax_id)

        table = Table(domain, data_x)
        table.name = 'Gene Matcher Results'
        table.attributes[OrangeTableAnnotations.tax_id] = next(tax_id)
        table.attributes[OrangeTableAnnotations.gene_as_attribute_name] = False
        table.attributes[OrangeTableAnnotations.gene_id_column] = NCBI_ID
        return table

    def run_matcher(self, progress_callback=None):
        """ This will try to match genes, with ncbi ids, based on provided input of genes.

        :param progress_callback: Used for progress bar in widgets. It emits the signal back to main thread

        """

        if progress_callback:
            self._match(callback=progress_callback.emit)
        else:
            self._match()

    def _match(self, **kwargs):
        # TODO: refactor this madness ...
        callback = kwargs.get("callback", None)

        def match_input(mapper, input_name):
            return mapper[input_name]

        for gene in self.genes:
            if callback:
                callback()

            try:
                ncbi_match = match_input(self._matcher[MAP_GENE_ID], int(gene.input_name))
                if ncbi_match:
                    gene.ncbi_id = ncbi_match[0][MAP_GENE_ID]
                    gene.type_of_match = _gene_id
                    gene.load_ncbi_info()
                    continue
            except ValueError:
                # NCBI ids are stored as Integers. If ValueError is raised, probably not NCBI ID.
                # We expect unique match here.
                pass

            input_name = gene.input_name
            if self._case_insensitive:
                input_name = gene.input_name.lower()

            source_match = match_input(self._matcher[MAP_SOURCES], gene.input_name)
            # ids from different sources are unique. We do not expect to get multiple hits here.
            # There is exceptions with organism 3702. It has same source id from Araport or TAIR databases.
            if source_match:
                gene.ncbi_id = source_match[0][MAP_GENE_ID]
                gene.type_of_match = _source
                gene.load_ncbi_info()
                continue

            symbol_match = match_input(self._matcher[MAP_SYMBOL], input_name)
            if len(symbol_match) == _single_hit:
                gene.ncbi_id = symbol_match[0][MAP_GENE_ID]
                gene.type_of_match = _symbol
                gene.load_ncbi_info()
                continue
            elif len(symbol_match) >= _multiple_hits:
                gene.possible_hits = symbol_match
                continue

            locus_match = match_input(self._matcher[MAP_LOCUS], gene.input_name)
            if len(locus_match) == _single_hit:
                gene.ncbi_id = locus_match[0][MAP_GENE_ID]
                gene.type_of_match = _locus
                gene.load_ncbi_info()
                continue
            elif len(symbol_match) >= _multiple_hits:
                gene.possible_hits = locus_match
                continue

            synonym_match = match_input(self._matcher[MAP_SYNONYMS], input_name)
            if len(synonym_match) == _single_hit:
                gene.ncbi_id = synonym_match[0][MAP_GENE_ID]
                gene.type_of_match = _synonym
                gene.load_ncbi_info()
                continue
            elif len(synonym_match) >= _multiple_hits:
                gene.possible_hits = synonym_match
                continue

            nomenclature_match = match_input(self._matcher[MAP_NOMENCLATURE], input_name)
            if len(nomenclature_match) == _single_hit:
                gene.ncbi_id = nomenclature_match[0][MAP_GENE_ID]
                gene.type_of_match = _nom_symbol
                gene.load_ncbi_info()
                continue
            elif len(nomenclature_match) >= _multiple_hits:
                gene.possible_hits = nomenclature_match
                continue

    def load_matcher_file(self, domain, filename):
        try:
            # this starts download if files are not on local machine
            file_path = serverfiles.localpath_download(domain, filename)
            # download new version before using this file for gene name matching
            # serverfiles.update(domain, filename)
        except (ConnectTimeout, RequestException, ConnectionError) as e:
            # Do not raise exception.
            print(e)
            return

        def case_insensitive_keys(matcher_dict):
            updated_dict = {MAP_SOURCES:  matcher_dict[MAP_SOURCES],
                            MAP_GENE_ID: matcher_dict[MAP_GENE_ID],
                            MAP_LOCUS:    matcher_dict[MAP_LOCUS],
                            MAP_SYNONYMS: defaultdict(list),
                            MAP_SYMBOL:  defaultdict(list),
                            MAP_NOMENCLATURE: defaultdict(list)}

            for key, value in matcher_dict[MAP_SYMBOL].items():
                # ensure string, we are using string methods (upper, lower)
                key = ensure_type(str(key), str)
                updated_dict[MAP_SYMBOL][key] = value
                updated_dict[MAP_SYMBOL][key.lower()] = value

            for key, value in matcher_dict[MAP_SYNONYMS].items():
                key = ensure_type(str(key), str)
                updated_dict[MAP_SYNONYMS][key] = value
                updated_dict[MAP_SYNONYMS][key.lower()] = value

            for key, value in matcher_dict[MAP_NOMENCLATURE].items():
                key = ensure_type(str(key), str)
                updated_dict[MAP_NOMENCLATURE][key] = value
                updated_dict[MAP_NOMENCLATURE][key.lower()] = value

            return updated_dict

        with open(file_path, 'rb') as pickle_file:
            if not self._case_insensitive:
                return pickle.load(pickle_file)
            else:
                return case_insensitive_keys(pickle.load(pickle_file))


if __name__ == "__main__":
    g = Gene()
    g.ncbi_id = 1
    g.load_ncbi_info()
    print(g.__slots__)
    print(g.ncbi_id, g.description, g.db_refs, g.synonyms)



""" OWMarkerGenes """
import sys
import numpy as np


from typing import List
from collections import defaultdict
from functools import partial


from AnyQt.QtCore import (
    Qt,
    QSize,
    QTimer,
    QModelIndex,
    QItemSelection,
    QItemSelectionModel,
    QItemSelectionRange,
)
from AnyQt.QtGui import QFont, QColor
from AnyQt.QtWidgets import QTreeView, QLineEdit

from Orange.data import MISSING_VALUES, Table, Domain
from Orange.widgets import widget, gui, settings
from Orange.widgets.utils.itemmodels import TableModel

from orangecontrib.bioinformatics.utils import serverfiles
from orangecontrib.bioinformatics.widgets.utils.data import (
    GENE_AS_ATTRIBUTE_NAME,
    TAX_ID,
    GENE_ID_COLUMN,
)

serverfiles_domain = 'marker_genes'


def get_available_db_sources():
    found_sources = dict()

    try:
        found_sources.update(
            serverfiles.ServerFiles().allinfo(serverfiles_domain)
        )
    except ConnectionError as e:
        raise ConnectionError(
            'Can not connect to {}. Using only local files.'.format(
                serverfiles.server_url
            )
        )
    finally:
        found_sources.update(serverfiles.allinfo(serverfiles_domain))
        return {item.get('title').split(': ')[-1]: item for item in found_sources.values()}


# todo: this is ugly, refactor this in the future.
class HeaderIndex:
    NAME = 0
    GENE = 1
    CELL_TYPE = 2
    FUNCTION = 3
    REFERENCE = 4
    URL = 5


HeaderLabels = {
    HeaderIndex.GENE: 'Entrez ID',
    HeaderIndex.REFERENCE: 'Reference',
    HeaderIndex.URL: 'URL',
}


class SearchableTableModel(TableModel):

    ClassVar, Meta, Attribute = range(3)

    ColorForRole = {ClassVar: None, Meta: None, Attribute: None}

    def __init__(self, data, parent):
        TableModel.__init__(self, data, parent)
        self._data = data
        self._roleData = {Qt.DisplayRole: self.source}
        self._roleData = partial(
            defaultdict,
            partial(defaultdict, partial(defaultdict, lambda: None)),
        )(self._roleData)
        self.set_column_links()

    def set_column_links(self):
        domain = self._data.domain
        ref_col = domain.metas.index(
            domain[HeaderLabels[HeaderIndex.REFERENCE]]
        )
        font = QFont()
        font.setUnderline(True)
        color = QColor(Qt.blue)
        for i, row in enumerate(self.source):
            link = row[HeaderLabels[HeaderIndex.URL]].value
            if len(link) and 'http' in link:
                self._roleData[gui.LinkRole][i][ref_col] = link
                self._roleData[Qt.FontRole][i][ref_col] = font
                self._roleData[Qt.ForegroundRole][i][ref_col] = color

    def data(self, index, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            cell_data = super().data(index, role)
            return "" if cell_data in MISSING_VALUES else cell_data
        elif role in (gui.LinkRole, Qt.FontRole, Qt.ForegroundRole):
            row, col = index.row(), index.column()
            return self._roleData[role][row][col]
        return super().data(index, role)

    def rowCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else len(self.source)

    def wrap(self, table):
        self.beginResetModel()
        self.source = table
        self._roleData.clear()
        self.resetSorting()
        self.endResetModel()

    def filter_table(self, filter_pattern):
        selection = np.full(self._data.metas.shape, True)
        for search_word in filter_pattern.split():
            match_result = (
                np.core.defchararray.find(
                    np.char.lower(self._data.metas.astype(str)),
                    search_word.lower(),
                )
                >= 0
            )
            selection = selection & match_result
        return selection

    def update_model(self, filter_pattern=None):
        # clear cache if model changes
        self._row_instance.cache_clear()
        self.wrap(self._data[self.filter_table(filter_pattern).any(axis=1), :])
        self.set_column_links()


class MarkerGroupContextHandler(settings.ContextHandler):
    def __init__(self):
        super().__init__()

    def match(self, context, group, *args):
        if not context.group == group:
            return self.NO_MATCH

        return self.PERFECT_MATCH

    def new_context(self, group):
        context = super().new_context()
        context.group = group
        return context


class OWMarkerGenes(widget.OWWidget):
    name = "Marker Genes"
    icon = 'icons/OWMarkerGenes.svg'
    priority = 170

    replaces = ['orangecontrib.single_cell.widgets.owmarkergenes.OWMarkerGenes']

    class Outputs:
        genes = widget.Output("Genes", Table)

    want_main_area = True

    selected_group: str = settings.Setting('')
    selected_db_source: str = settings.Setting('')
    filter_text: str = settings.Setting('')
    header_state: bytes = settings.Setting(b'')

    settingsHandler = MarkerGroupContextHandler()
    selected_genes: List[tuple] = settings.ContextSetting([])

    def __init__(self):
        super().__init__()
        self._data = None
        self._available_db_sources = None

        self._timer = QTimer()
        self._timer.timeout.connect(self._filter_table)
        self._timer.setSingleShot(True)

        box = gui.widgetBox(self.controlArea, 'Database', margin=0)
        self.db_source_index = -1
        self.db_source_cb = gui.comboBox(box, self, 'db_source_index')
        self.db_source_cb.activated[int].connect(self.handle_source_changed)

        box = gui.widgetBox(self.controlArea, 'Organism', margin=0)
        self.group_index = -1
        self.group_cb = gui.comboBox(box, self, 'group_index')
        self.group_cb.activated[int].connect(self.set_group_index)
        gui.rubber(self.controlArea)

        # TODO: to avoid this, marker genes table should have 'tax_id' column
        self.map_group_to_taxid = {'Human': '9606', 'Mouse': '10090'}

        filter_line_edit = gui.lineEdit(
            self.mainArea, self, "filter_text"
        )  # type: QLineEdit
        filter_line_edit.setPlaceholderText("Filter...")
        filter_line_edit.textEdited.connect(self.call_filter_timer)

        self.view = view = QTreeView(
            rootIsDecorated=False,
            uniformRowHeights=True,
            selectionMode=QTreeView.ExtendedSelection,
            sortingEnabled=True,
        )

        view.viewport().setMouseTracking(True)
        self.mainArea.layout().addWidget(view)

        self._load_data()
        if self.header_state:
            view.header().restoreState(self.header_state)

    @property
    def available_db_sources(self) -> dict:
        return self._available_db_sources

    @available_db_sources.setter
    def available_db_sources(self, value: dict):
        self._available_db_sources = value

        items = list(value.keys())
        try:
            idx = items.index(self.selected_db_source)
        except ValueError:
            idx = -1

        self.db_source_cb.clear()
        self.db_source_cb.addItems(items)

        if idx != -1:
            self.db_source_index = idx
            self.selected_db_source = items[idx]
        elif items:
            self.db_source_index = min(
                max(self.db_source_index, 0), len(items) - 1
            )

        self.set_db_source_index(self.db_source_index)

    @property
    def data(self) -> Table:
        return self._data

    @data.setter
    def data(self, value: Table):
        """ Set the source data.

        The data is then filtered on the first meta column (group)
        """
        self._data = value
        domain = value.domain

        if domain.metas:
            group = domain.metas[0]
            groupcol, _ = value.get_column_view(group)

            if group.is_string:
                group_values = list(set(groupcol))
            elif group.is_discrete:
                group_values = group.values
            else:
                raise TypeError("Invalid column type")
            try:
                idx = group_values.index(self.selected_group)
            except ValueError:
                idx = -1

            self.group_cb.clear()
            self.group_cb.addItems(group_values)

            if idx != -1:
                self.group_index = idx
                self.selected_group = group_values[idx]
            elif group_values:
                self.group_index = min(
                    max(self.group_index, 0), len(group_values) - 1
                )

            self.set_group_index(self.group_index)

    def _load_data(self):
        self.available_db_sources = get_available_db_sources()
        file_name = self.available_db_sources[self.selected_db_source][
            'filename'
        ]

        try:
            serverfiles.update(serverfiles_domain, file_name)
        except ConnectionError as e:
            raise ConnectionError(
                'Can not connect to {}. '
                'Using only local files.'.format(serverfiles.server_url)
            )
        finally:
            file_path = serverfiles.localpath_download(
                serverfiles_domain, file_name
            )
            data = Table(file_path)

            # enforce order
            old_domain = data.domain
            new_domain = Domain(
                [],
                metas=[
                    old_domain['Organism'],
                    old_domain['Name'],
                    old_domain['Entrez ID'],
                    old_domain['Cell Type'],
                    old_domain['Function'],
                    old_domain['Reference'],
                    old_domain['URL'],
                ],
            )
            data = data.transform(new_domain)
            self.data = data

    def set_selection(self):
        selected = self.selected_rows()

        if len(selected):
            header_count = self.view.header().count() - 1

            if self.view.model().rowCount() <= selected[-1]:
                return

            selection = QItemSelection()

            for row_index in selected:
                selection.append(
                    QItemSelectionRange(
                        self.view.model().index(row_index, 0),
                        self.view.model().index(row_index, header_count),
                    )
                )

            self.view.selectionModel().select(
                selection, QItemSelectionModel.ClearAndSelect
            )

    def handle_source_changed(self, source_index):
        self.set_db_source_index(source_index)
        self._load_data()

    def set_db_source_index(self, source_index):
        self.closeContext()
        self.db_source_index = source_index
        self.selected_db_source = self.db_source_cb.itemText(source_index)

    def set_group_index(self, group_index):
        self.closeContext()
        self.group_index = group_index
        self.selected_group = self.group_cb.itemText(group_index)
        self._setup()

    def call_filter_timer(self, search_string):
        self._timer.stop()
        if search_string != self.filter_text:
            self.filter_text = search_string
        self._timer.start(500)

    def _filter_table(self):
        model = self.view.model()
        assert isinstance(model, SearchableTableModel)
        model.update_model(str(self.filter_text))

    def _setup(self):
        self.closeContext()
        data = self.data
        group = data.domain.metas[0]
        gvec = data.get_column_view(group)[0]

        if group.is_string:
            mask = gvec == self.group_cb.itemData(
                self.group_index, Qt.DisplayRole
            )
        else:
            mask = gvec == self.group_index

        data = data[mask]
        rest = data[:, data.domain.metas[1:]]
        model = SearchableTableModel(rest, parent=self)
        ref_col = rest.domain.metas.index(
            rest.domain[HeaderLabels[HeaderIndex.REFERENCE]]
        )
        self.view.setItemDelegateForColumn(
            ref_col, gui.LinkStyledItemDelegate(self.view)
        )

        self.view.setModel(model)
        self.view.selectionModel().selectionChanged.connect(
            self._on_selection_changed
        )

        self.openContext(self.selected_group)
        self.call_filter_timer(self.filter_text)
        self.view.hideColumn(HeaderIndex.URL)
        self.set_selection()

        self.commit()

    def _on_selection_changed(self):
        self.commit()

    def selected_rows(self):
        """ Return row index for selected genes
        """
        if not self.selected_genes:
            return []

        model = self.view.model()
        return [
            row_index
            for row_index in range(model.rowCount())
            if (
                model.index(row_index, HeaderIndex.GENE).data(),
                model.index(row_index, HeaderIndex.CELL_TYPE).data(),
                model.index(row_index, HeaderIndex.REFERENCE).data(),
            )
            in self.selected_genes
        ]

    def commit(self):
        model = self.view.model()
        assert isinstance(model, SearchableTableModel)
        rows = [mi.row() for mi in self.view.selectionModel().selectedRows(0)]

        if rows:
            rows = model.mapToSourceRows(rows)
            output = model.source[rows]
        else:
            output = model.source

        gene_id = self.view.selectionModel().selectedRows(HeaderIndex.GENE)
        cell_type = self.view.selectionModel().selectedRows(
            HeaderIndex.CELL_TYPE
        )
        ref = self.view.selectionModel().selectedRows(HeaderIndex.REFERENCE)

        self.selected_genes = [
            (entrez.data(), cell.data(), ref.data())
            for entrez, cell, ref in zip(gene_id, cell_type, ref)
        ]

        # always false for marker genes data tables in single cell
        output.attributes[GENE_AS_ATTRIBUTE_NAME] = False
        # set taxonomy id in data.attributes
        output.attributes[TAX_ID] = self.map_group_to_taxid.get(
            self.selected_group, ''
        )
        # set column id flag
        output.attributes[GENE_ID_COLUMN] = HeaderLabels[HeaderIndex.GENE]
        output.name = 'Marker Genes'

        self.Outputs.genes.send(output)

    def closeEvent(self, event):
        self.header_state = bytes(self.view.header().saveState())
        super().closeEvent(event)

    def sizeHint(self):
        return super().sizeHint().expandedTo(QSize(600, 500))


def main(argv=None):
    from AnyQt.QtWidgets import QApplication

    app = QApplication(argv or sys.argv)
    w = OWMarkerGenes()
    w.show()
    w.activateWindow()
    rv = app.exec_()
    w.saveSettings()
    w.onDeleteWidget()
    return rv


if __name__ == "__main__":
    sys.exit(main())
